import itertools
import logging
import os
import re
import string
from collections import defaultdict
from collections.abc import Sequence
from dataclasses import dataclass
from enum import Enum
from functools import cache, cached_property, total_ordering
from typing import Dict, List, Optional, Tuple

import graphviz
import pulp

LOGLEVEL = os.environ.get("LOGLEVEL", "INFO").upper()
logging.basicConfig(level=LOGLEVEL)


class Molecule(Enum):
    DNA = "DNA"
    RNA = "RNA"
    Other = "Other"


class GlycosidicBond(Enum):
    anti = "anti"
    syn = "syn"


@total_ordering
class LeontisWesthof(Enum):
    cWW = "cWW"
    cWH = "cWH"
    cWS = "cWS"
    cHW = "cHW"
    cHH = "cHH"
    cHS = "cHS"
    cSW = "cSW"
    cSH = "cSH"
    cSS = "cSS"
    tWW = "tWW"
    tWH = "tWH"
    tWS = "tWS"
    tHW = "tHW"
    tHH = "tHH"
    tHS = "tHS"
    tSW = "tSW"
    tSH = "tSH"
    tSS = "tSS"

    @property
    def reverse(self):
        return LeontisWesthof[f"{self.name[0]}{self.name[2]}{self.name[1]}"]

    def __lt__(self, other):
        return tuple(self.value) < tuple(other.value)


class Saenger(Enum):
    I = "I"
    II = "II"
    III = "III"
    IV = "IV"
    V = "V"
    VI = "VI"
    VII = "VII"
    VIII = "VIII"
    IX = "IX"
    X = "X"
    XI = "XI"
    XII = "XII"
    XIII = "XIII"
    XIV = "XIV"
    XV = "XV"
    XVI = "XVI"
    XVII = "XVII"
    XVIII = "XVIII"
    XIX = "XIX"
    XX = "XX"
    XXI = "XXI"
    XXII = "XXII"
    XXIII = "XXIII"
    XXIV = "XXIV"
    XXV = "XXV"
    XXVI = "XXVI"
    XXVII = "XXVII"
    XXVIII = "XXVIII"

    @staticmethod
    def table() -> Dict[Tuple[str, str], str]:
        return {
            ("AA", "tWW"): "I",
            ("AA", "tHH"): "II",
            ("GG", "tWW"): "III",
            ("GG", "tSS"): "IV",
            ("AA", "tWH"): "V",
            ("AA", "tHW"): "V",
            ("GG", "cWH"): "VI",
            ("GG", "cHW"): "VI",
            ("GG", "tWH"): "VII",
            ("GG", "tHW"): "VII",
            ("AG", "cWW"): "VIII",
            ("GA", "cWW"): "VIII",
            ("AG", "cHW"): "IX",
            ("GA", "cWH"): "IX",
            ("AG", "tWS"): "X",
            ("GA", "tSW"): "X",
            ("AG", "tHS"): "XI",
            ("GA", "tSH"): "XI",
            ("UU", "tWW"): "XII",
            ("TT", "tWW"): "XII",
            # XIII is UU/TT in tWW but donor-donor, so impossible
            # XIV and XV are both CC in tWW but donor-donor, so impossible
            ("UU", "cWW"): "XVI",
            ("TT", "cWW"): "XVI",
            ("CU", "tWW"): "XVII",
            ("UC", "tWW"): "XVII",
            ("CU", "cWW"): "XVIII",
            ("UC", "cWW"): "XVIII",
            ("CG", "cWW"): "XIX",
            ("GC", "cWW"): "XIX",
            ("AU", "cWW"): "XX",
            ("UA", "cWW"): "XX",
            ("AT", "cWW"): "XX",
            ("TA", "cWW"): "XX",
            ("AU", "tWW"): "XXI",
            ("UA", "tWW"): "XXI",
            ("AT", "tWW"): "XXI",
            ("TA", "tWW"): "XXI",
            ("CG", "tWW"): "XXII",
            ("GC", "tWW"): "XXII",
            ("AU", "cHW"): "XXIII",
            ("UA", "cWH"): "XXIII",
            ("AT", "cHW"): "XXIII",
            ("TA", "cWH"): "XXIII",
            ("AU", "tHW"): "XXIV",
            ("UA", "tWH"): "XXIV",
            ("AT", "tHW"): "XXIV",
            ("TA", "tWH"): "XXIV",
            ("AC", "tHW"): "XXV",
            ("CA", "tWH"): "XXV",
            ("AC", "tWW"): "XXVI",
            ("CA", "tWW"): "XXVI",
            ("GU", "tWW"): "XXVII",
            ("UG", "tWW"): "XXVII",
            ("GT", "tWW"): "XXVII",
            ("TG", "tWW"): "XXVII",
            ("GU", "cWW"): "XXVIII",
            ("UG", "cWW"): "XXVIII",
            ("GT", "cWW"): "XXVIII",
            ("TG", "cWW"): "XXVIII",
        }

    @property
    def is_canonical(self) -> bool:
        return self == Saenger.XIX or self == Saenger.XX or self == Saenger.XXVIII


class StackingTopology(Enum):
    upward = "upward"
    downward = "downward"
    inward = "inward"
    outward = "outward"

    @property
    def reverse(self):
        if self == StackingTopology.upward:
            return StackingTopology.downward
        elif self == StackingTopology.downward:
            return StackingTopology.upward
        return self


class BR(Enum):
    _0 = "0BR"
    _1 = "1BR"
    _2 = "2BR"
    _3 = "3BR"
    _4 = "4BR"
    _5 = "5BR"
    _6 = "6BR"
    _7 = "7BR"
    _8 = "8BR"
    _9 = "9BR"


class BPh(Enum):
    _0 = "0BPh"
    _1 = "1BPh"
    _2 = "2BPh"
    _3 = "3BPh"
    _4 = "4BPh"
    _5 = "5BPh"
    _6 = "6BPh"
    _7 = "7BPh"
    _8 = "8BPh"
    _9 = "9BPh"


@dataclass(frozen=True, order=True)
class ResidueLabel:
    chain: str
    number: int
    name: str


@dataclass(frozen=True, order=True)
class ResidueAuth:
    chain: str
    number: int
    icode: Optional[str]
    name: str


@dataclass(frozen=True)
@total_ordering
class Residue:
    label: Optional[ResidueLabel]
    auth: Optional[ResidueAuth]

    def __lt__(self, other):
        return (self.chain, self.number, self.icode or " ") < (
            other.chain,
            other.number,
            other.icode or " ",
        )

    @property
    def chain(self) -> Optional[str]:
        if self.auth is not None:
            return self.auth.chain
        if self.label is not None:
            return self.label.chain
        return None

    @property
    def number(self) -> Optional[int]:
        if self.auth is not None:
            return self.auth.number
        if self.label is not None:
            return self.label.number
        return None

    @property
    def icode(self) -> Optional[str]:
        if self.auth is not None:
            return self.auth.icode if self.auth.icode not in (" ", "?") else None
        return None

    @property
    def name(self) -> Optional[str]:
        if self.auth is not None:
            return self.auth.name
        if self.label is not None:
            return self.label.name
        return None

    @property
    def molecule_type(self) -> Molecule:
        if self.name is not None:
            if self.name.upper() in ("A", "C", "G", "U"):
                return Molecule.RNA
            if self.name.upper() in ("DA", "DC", "DG", "DT"):
                return Molecule.DNA
        return Molecule.Other

    @property
    @cache
    def full_name(self) -> Optional[str]:
        if self.auth is not None:
            if self.auth.chain.isspace():
                builder = f"{self.auth.name}"
            else:
                builder = f"{self.auth.chain}.{self.auth.name}"
            if len(self.auth.name) > 0 and self.auth.name[-1] in string.digits:
                builder += "/"
            builder += f"{self.auth.number}"
            if self.auth.icode:
                builder += f"^{self.auth.icode}"
            return builder
        elif self.label is not None:
            if self.label.chain.isspace():
                builder = f"{self.label.name}"
            else:
                builder = f"{self.label.chain}.{self.label.name}"
            if len(self.label.name) > 0 and self.label.name[-1] in string.digits:
                builder += "/"
            builder += f"{self.label.number}"
            return builder
        return None


@dataclass(frozen=True, order=True)
class Interaction:
    nt1: Residue
    nt2: Residue


@dataclass(frozen=True, order=True)
class BasePair(Interaction):
    lw: LeontisWesthof
    saenger: Optional[Saenger]


@dataclass(frozen=True, order=True)
class Stacking(Interaction):
    topology: Optional[StackingTopology]


@dataclass(frozen=True, order=True)
class BaseRibose(Interaction):
    br: Optional[BR]


@dataclass(frozen=True, order=True)
class BasePhosphate(Interaction):
    bph: Optional[BPh]


@dataclass(frozen=True, order=True)
class OtherInteraction(Interaction):
    pass


@dataclass
class Entry(Sequence):
    index_: int
    sequence: str
    pair: int

    def __getitem__(self, item):
        if item == 0:
            return self.index_
        elif item == 1:
            return self.sequence
        elif item == 2:
            return self.pair
        raise IndexError()

    def __lt__(self, other):
        return self.index_ < other.index_

    def __len__(self) -> int:
        return 3

    def __str__(self):
        return f"{self.index_} {self.sequence} {self.pair}"


@dataclass(frozen=True)
class Strand:
    first: int
    last: int
    sequence: str
    structure: str

    @staticmethod
    def from_bpseq_entries(
        entries: List[Entry], dotbracket: str, reverse: bool = False
    ):
        first = entries[0].index_
        last = first + len(entries) - 1
        if reverse:
            first, last = last, first
        sequence = "".join(
            [
                entry.sequence
                for entry in (entries if not reverse else reversed(entries))
            ]
        )
        structure = dotbracket[first - 1 : last]
        return Strand(first, last, sequence, structure)

    def __str__(self):
        return f"{self.first}-{self.sequence}-{self.last}"


@dataclass
class SingleStrand:
    strand: Strand
    is5p: bool
    is3p: bool

    def __post_init__(self):
        self.description = str(self)

    def __str__(self):
        if self.is5p:
            return f"SingleStrand5p {self.strand.first} {self.strand.last} {self.strand.sequence} {self.strand.structure}"
        if self.is3p:
            return f"SingleStrand3p {self.strand.first} {self.strand.last} {self.strand.sequence} {self.strand.structure}"
        return f"SingleStrand {self.strand.first} {self.strand.last} {self.strand.sequence} {self.strand.structure}"


@dataclass
class Stem:
    strand5p: Strand
    strand3p: Strand

    @staticmethod
    def from_bpseq_entries(
        strand5p_entries: List[Entry], all_entries: List, dotbracket: str
    ):
        paired = set([entry[2] for entry in strand5p_entries])
        strand3p_entries = list(filter(lambda entry: entry[0] in paired, all_entries))
        return Stem(
            Strand.from_bpseq_entries(strand5p_entries, dotbracket),
            Strand.from_bpseq_entries(strand3p_entries, dotbracket),
        )

    def __post_init__(self):
        self.description = str(self)

    def __str__(self):
        return f"Stem {self.strand5p.first} {self.strand5p.last} {self.strand5p.sequence} {self.strand5p.structure} {self.strand3p.first} {self.strand3p.last} {self.strand3p.sequence} {self.strand3p.structure}"


@dataclass
class Hairpin:
    strand: Strand

    def __post_init__(self):
        self.description = str(self)

    def __str__(self):
        return f"Hairpin {self.strand.first} {self.strand.last} {self.strand.sequence} {self.strand.structure}"


@dataclass
class Loop:
    strands: List[Strand]

    def __post_init__(self):
        self.description = str(self)

    def __str__(self):
        desc = " ".join(
            [
                "{} {} {} {}".format(
                    strand.first, strand.last, strand.sequence, strand.structure
                )
                for strand in self.strands
            ]
        )
        return f"Loop {desc}"


@dataclass
class BpSeq:
    entries: List[Entry]

    @staticmethod
    def from_string(bpseq_str: str):
        entries = []
        for line in bpseq_str.splitlines():
            line = line.strip()
            if len(line) == 0:
                continue
            fields = line.split()
            if len(fields) != 3:
                logging.warning("Failed to find 3 columns in BpSeq line: {}", line)
                continue
            entry = Entry(int(fields[0]), fields[1], int(fields[2]))
            entries.append(entry)
        return BpSeq(entries)

    @staticmethod
    def from_file(bpseq_path: str):
        with open(bpseq_path) as f:
            return BpSeq.from_string(f.read())

    @staticmethod
    def from_dotbracket(dot_bracket):
        entries = [
            Entry(i + 1, dot_bracket.sequence[i], 0)
            for i in range(len(dot_bracket.sequence))
        ]
        for i, j in dot_bracket.pairs:
            entries[i].pair = j + 1
            entries[j].pair = i + 1
        return BpSeq(entries)

    def __post_init__(self):
        self.pairs = {}
        for i, _, j in self.entries:
            if j != 0:
                self.pairs[i] = j
                self.pairs[j] = i

    def __str__(self):
        return "\n".join(("{} {} {}".format(i, c, j) for i, c, j in self.entries))

    def __eq__(self, other):
        return len(self.entries) == len(other.entries) and all(
            ei == ej for ei, ej in zip(self.entries, other.entries)
        )

    @cached_property
    def sequence(self) -> str:
        return "".join(entry.sequence for entry in self.entries)

    def paired(self, only5to3: bool = False):
        result = filter(lambda entry: entry.pair != 0, self.entries)
        if only5to3:
            result = filter(lambda entry: entry.index_ < entry.pair, result)
        return result

    @cached_property
    def __stems_entries(self) -> List[List[Entry]]:
        stems = []
        entries: List[Entry] = []

        for entry in self.paired(only5to3=True):
            if not entries:
                entries.append(entry)
                continue

            i, _, j = entry
            k, _, l = entries[-1]
            if i == k + 1 and j == l - 1:
                entries.append(entry)
                continue

            stems.append(entries)
            entries = [entry]

        if entries:
            stems.append(entries)

        return stems

    @cached_property
    def elements(
        self,
    ) -> Tuple[List[Stem], List[SingleStrand], List[Hairpin], List[Loop]]:
        if not self.__stems_entries:
            return [], [], [], []

        stems, single_strands, hairpins, loops = [], [], [], []
        stopset = set()

        # stems
        for stem_entries in self.__stems_entries:
            stem = Stem.from_bpseq_entries(
                stem_entries, self.entries, self.dot_bracket.structure
            )
            stems.append(stem)
            stopset.add(stem.strand5p.first - 1)
            stopset.add(stem.strand5p.last - 1)
            stopset.add(stem.strand3p.first - 1)
            stopset.add(stem.strand3p.last - 1)

        stops = sorted(stopset)
        loop_candidates = []

        # 5' single strand
        if stops[0] > 0:
            single_strands.append(
                SingleStrand(
                    Strand.from_bpseq_entries(
                        self.entries[: stops[0] + 1],
                        self.dot_bracket.structure,
                    ),
                    True,
                    False,
                )
            )

        # single strands
        for i in range(1, len(stops)):
            candidate = self.entries[stops[i - 1] : stops[i] + 1]
            if all([entry.pair == 0 for entry in candidate[1:-1]]):
                if candidate[0].pair == candidate[-1].index_:
                    hairpins.append(
                        Hairpin(
                            Strand.from_bpseq_entries(
                                candidate, self.dot_bracket.structure
                            )
                        )
                    )
                else:
                    loop_candidates.append(
                        Strand.from_bpseq_entries(candidate, self.dot_bracket.structure)
                    )

        # 3' single strand
        if stops[-1] < len(self.entries) - 1:
            single_strands.append(
                SingleStrand(
                    Strand.from_bpseq_entries(
                        self.entries[stops[-1] :], self.dot_bracket.structure
                    ),
                    False,
                    True,
                )
            )

        graph = defaultdict(set)

        for i in range(len(loop_candidates)):
            for j in range(i + 1, len(loop_candidates)):
                i_first, i_last = loop_candidates[i].first, loop_candidates[i].last
                j_first, j_last = loop_candidates[j].first, loop_candidates[j].last
                if self.entries[i_last - 1].pair == j_first:
                    graph[i].add(j)
                if self.entries[j_last - 1].pair == i_first:
                    graph[j].add(i)

        used = set()

        for i in range(len(loop_candidates)):
            if i in used:
                continue

            loop = [loop_candidates[i]]

            while True:
                for j in graph[i]:
                    if (
                        loop_candidates[j] not in used
                        and loop_candidates[j] not in loop
                    ):
                        loop.append(loop_candidates[j])
                        i = j
                        break
                else:
                    break

            if self.entries[loop[0].first - 1].pair == loop[-1].last:
                if not all([strand.last - strand.first <= 1 for strand in loop]):
                    loops.append(Loop(loop))
                    used.update(loop)

        for loop_candidate in loop_candidates:
            if loop_candidate not in used:
                single_strands.append(SingleStrand(loop_candidate, False, False))

        return stems, single_strands, hairpins, loops

    @cached_property
    def graphviz(self):
        stems, single_strands, hairpins, loops = self.elements
        graph = defaultdict(set)
        dot = graphviz.Graph()

        for single_strand in single_strands:
            graph[str(single_strand)].update(
                [
                    single_strand.strand.first,
                    single_strand.strand.last,
                ]
            )

        for stem in stems:
            if stem.strand5p.first == stem.strand5p.last:
                continue
            graph[str(stem)].update(
                [
                    stem.strand5p.first,
                    stem.strand5p.last,
                    stem.strand3p.first,
                    stem.strand3p.last,
                ]
            )

        for hairpin in hairpins:
            graph[str(hairpin)].update(
                [
                    hairpin.strand.first,
                    hairpin.strand.last,
                ]
            )

        for loop in loops:
            stops = set()
            for strand in loop.strands:
                stops.update(
                    [
                        strand.first,
                        strand.last,
                    ]
                )
            graph[str(loop)].update(stops)

        for i, element in enumerate(graph.keys()):
            dot.node(f"E{i}", str(element))

        keys = list(graph.keys())

        for i in range(len(keys)):
            for j in range(i + 1, len(keys)):
                if graph[keys[i]].intersection(graph[keys[j]]):
                    dot.edge(f"E{i}", f"E{j}")

        return dot.render()

    @cached_property
    def __regions(self) -> List[Tuple[int, int, int]]:
        return [
            (stem_entries[0].index_, stem_entries[0].pair, len(stem_entries))
            for stem_entries in self.__stems_entries
        ]

    @cached_property
    def dot_bracket(self):
        if pulp.HiGHS_CMD().available():
            solver = pulp.HiGHS_CMD()  # much faster than default
        else:
            solver = pulp.LpSolverDefault
        if solver is not None:
            solver.msg = False
        return self.convert_to_dot_bracket(solver)

    def convert_to_dot_bracket(self, solver: pulp.LpSolver):
        # if PuLP solvers are not installed, use FCFS
        if solver is None:
            return self.fcfs()

        # build conflict graph
        regions = self.__regions
        graph = defaultdict(set)

        for i, j in itertools.combinations(range(len(regions)), 2):
            ri, rj = regions[i], regions[j]
            k, l, _ = ri
            m, n, _ = rj

            # is pseudoknot?
            if (k < m < l < n) or (m < k < n < l):
                graph[i].add(j)
                graph[j].add(i)

        # return all non-pseudoknotted if the graph is empty
        if not graph:
            return self.__make_dot_bracket(regions, [0 for _ in range(len(regions))])

        # determine maximum pseudoknot order as chromatic number bound equal to maximum vertex degree + 1
        max_order = max(map(len, graph.values())) + 1

        # define the problem
        problem = pulp.LpProblem("POA", pulp.LpMaximize)

        # create decision variables
        variables = []
        vars_by_region = defaultdict(list)
        vars_by_order = defaultdict(list)
        var_by_region_order = {}
        region_by_var = {}
        for i in range(len(regions)):
            for j in range(max_order):
                variable = pulp.LpVariable(f"x_{i}_{j}", 0, 1, pulp.LpInteger)
                variables.append(variable)
                vars_by_region[i].append(variable)
                vars_by_order[j].append(variable)
                var_by_region_order[(i, j)] = variable
                region_by_var[variable] = regions[i]

        # define objective function terms
        terms = []

        for order, vars in vars_by_order.items():
            for var in vars:
                length = region_by_var[var][2]
                if order == 0:
                    terms.append(var * length)
                else:
                    terms.append(-1 * var * length * order)

        # define objective function
        problem += pulp.lpSum(terms)

        # define constraints that each region is assigned to exactly one order
        for region_vars in vars_by_region.values():
            problem += pulp.lpSum(region_vars) == 1

        # define constraints that no two adjacent regions are assigned to the same order
        for i in graph.keys():
            for j in graph[i]:
                for order in range(max_order):
                    problem += (
                        var_by_region_order[(i, order)]
                        + var_by_region_order[(j, order)]
                        <= 1
                    )

        # solve the problem
        try:
            logging.debug(f"POA: problem formulation\n{problem}")
            problem.solve(solver)
        except pulp.PulpSolverError:
            logging.warning(
                "POA: failed to solve problem using MILP approach, fallback to FCFS"
            )
            return self.fcfs()

        # if problem is infeasible, fallback to FCFS
        if problem.status != pulp.LpStatusOptimal:
            logging.warning("POA: problem is infeasible, fallback to FCFS")
            return self.fcfs()

        # log solver time statistics
        logging.debug(
            f"POA: solver {solver.name} took {round(problem.solutionTime, 2)} seconds"
        )

        # map variable values to orders
        orders = [0 for _ in range(len(regions))]
        for variable in problem.variables():
            if variable.varValue == 1:
                name = variable.getName()
                i, order = map(int, name.split("_")[1:])
                orders[i] = order

        return self.__make_dot_bracket(regions, orders)

    def __make_dot_bracket(self, regions, orders):
        # build dot-bracket
        sequence = self.sequence
        structure = ["." for _ in range(len(sequence))]
        brackets = ["()", "[]", "{}", "<>"] + [
            "".join(p) for p in zip(string.ascii_uppercase, string.ascii_lowercase)
        ]

        for i, stem in enumerate(regions):
            bracket = brackets[orders[i]]
            j, k, n = stem

            while n > 0:
                structure[j - 1] = bracket[0]
                structure[k - 1] = bracket[1]
                j += 1
                k -= 1
                n -= 1

        structure = "".join(structure)
        return DotBracket.from_string(sequence, structure)

    @cached_property
    def fcfs(self):
        regions = [
            (stem_entries[0].index_, stem_entries[0].pair, len(stem_entries))
            for stem_entries in self.__stems_entries
        ]
        orders = [0 for i in range(len(regions))]

        for i in range(1, len(regions)):
            k, l, _ = regions[i]
            available = [True for _ in range(len("([{<" + string.ascii_uppercase))]

            for j in range(i):
                m, n, _ = regions[j]
                conflicted = (k < m < l < n) or (m < k < n < l)

                if conflicted:
                    available[orders[j]] = False

            order = next(filter(lambda i: available[i] is True, range(len(available))))
            orders[i] = order

        return self.__make_dot_bracket(regions, orders)

    @cached_property
    def all_dot_brackets(self):
        # build conflict graph
        regions = self.__regions
        graph = defaultdict(set)

        for i, j in itertools.combinations(range(len(regions)), 2):
            ri, rj = regions[i], regions[j]
            k, l, _ = ri
            m, n, _ = rj

            # is pseudoknot?
            if (k < m < l < n) or (m < k < n < l):
                graph[i].add(j)
                graph[j].add(i)

        # early exit for non-pseudoknotted structures
        vertices = list(graph.keys())
        if not vertices:
            return [self.fcfs]

        # find all connected components
        visited = {vertex: False for vertex in vertices}
        components = []

        for vertex in vertices:
            if not visited[vertex]:
                visited[vertex] = True
                stack = [vertex]
                components.append([vertex])

                while stack:
                    current = stack[-1]
                    next_vertex = None

                    for neighbor in graph[current]:
                        if not visited[neighbor]:
                            next_vertex = neighbor
                            break

                    if next_vertex is not None:
                        visited[next_vertex] = True
                        stack.append(next_vertex)
                        components[-1].append(next_vertex)
                    else:
                        stack.pop()

        # find unique orders for each component
        unique = []
        for component in components:
            unique.append(set())

            for permutation in itertools.permutations(component):
                orders = {region: 0 for region in component}

                for i in range(1, len(permutation)):
                    available = [True for _ in range(len(component))]

                    for j in range(i):
                        if permutation[j] in graph[permutation[i]]:
                            available[orders[permutation[j]]] = False

                    order = next(
                        filter(lambda k: available[k] is True, range(len(available)))
                    )
                    orders[permutation[i]] = order

                unique[-1].add(frozenset(orders.items()))

        # generate all possible dot-brackets
        solutions = set()
        for assignment in itertools.product(*unique):
            orders = {region: 0 for region in range(len(regions))}

            for order in assignment:
                orders.update(order)

            solutions.add(self.__make_dot_bracket(regions, orders))
        return list(solutions)

    def without_pseudoknots(self):
        return BpSeq.from_dotbracket(self.dot_bracket.without_pseudoknots())

    def without_isolated(self):
        stems, _, _, _ = self.elements
        to_unpair = []

        for stem in stems:
            if stem.strand5p.first == stem.strand5p.last:
                to_unpair.append(stem.strand5p.first - 1)
                to_unpair.append(stem.strand3p.first - 1)

        if not to_unpair:
            return self

        entries = self.entries.copy()
        for i in to_unpair:
            entries[i].pair = 0

        return BpSeq(entries)


@dataclass
class DotBracket:
    sequence: str
    structure: str

    @staticmethod
    def from_file(path: str):
        with open(path) as f:
            lines = f.readlines()
        if len(lines) == 2:
            return DotBracket.from_string(lines[0].rstrip(), lines[1].rstrip())
        if len(lines) == 3:
            return DotBracket.from_string(lines[1].rstrip(), lines[2].rstrip())
        raise RuntimeError(f"Failed to read DotBracket from file: {path}")

    @staticmethod
    def from_string(sequence: str, structure: str):
        if len(sequence) != len(structure):
            raise ValueError(
                "Sequence and structure lengths differ, {} vs {}",
                (len(sequence), len(structure)),
            )
        return DotBracket(sequence, structure)

    def __post_init__(self):
        self.pairs = []

        opening = "([{<" + string.ascii_uppercase
        closing = ")]}>" + string.ascii_lowercase
        begins = {bracket: list() for bracket in opening}
        matches = {end: begin for begin, end in zip(opening, closing)}

        for i in range(len(self.structure)):
            c = self.structure[i]
            if c in opening:
                begins[c].append(i)
            elif c in closing:
                begin = matches[c]
                self.pairs.append((begins[begin].pop(), i))

    def __str__(self):
        return f"{self.sequence}\n{self.structure}"

    def __eq__(self, other):
        return self.sequence == other.sequence and self.structure == other.structure

    def __hash__(self) -> int:
        return hash((self.sequence, self.structure))

    def without_pseudoknots(self):
        structure = re.sub(r"[\[\]\{\}\<\>A-Za-z]", ".", self.structure)
        return DotBracket(self.sequence, structure)


@dataclass
class MultiStrandDotBracket(DotBracket):
    strands: List[Strand]

    @staticmethod
    def from_string(sequence: str, structure: str):
        # Provide compatibility with DotBracket.from_string
        strand = Strand(1, len(sequence), sequence, structure)
        return MultiStrandDotBracket(sequence, structure, [strand])

    @staticmethod
    def from_multiline_string(input: str):
        strands = []
        first = 1

        for match in re.finditer(
            r"((>.*?\n)?([ACGTURYSWKMBDHVNacgturyswkmbdhvn.-]+)\n([.()\[\]{}<>A-Za-z]+))",
            input,
        ):
            sequence = match.group(3)
            structure = match.group(4)
            assert len(sequence) == len(structure)
            last = first + len(sequence) - 1
            strands.append(Strand(first, last, sequence, structure))
            first = last + 1

        return MultiStrandDotBracket(
            "".join(strand.sequence for strand in strands),
            "".join(strand.structure for strand in strands),
            strands,
        )

    @staticmethod
    def from_file(path: str):
        with open(path) as f:
            return MultiStrandDotBracket.from_multiline_string(f.read())


@dataclass(frozen=True, order=True)
class BaseInteractions:
    base_pairs: List[BasePair]
    stackings: List[Stacking]
    base_ribose_interactions: List[BaseRibose]
    base_phosphate_interactions: List[BasePhosphate]
    other_interactions: List[OtherInteraction]


@dataclass(frozen=True, order=True)
class InterStemParameters:
    stem1_idx: int
    stem2_idx: int
    type: Optional[str]  # Type of closest endpoint pair ('cs55', 'cs53', etc.)
    torsion: Optional[float]  # Torsion angle between stem segments (degrees)
    min_endpoint_distance: Optional[float]  # Minimum distance between stem endpoints
    torsion_angle_pdf: Optional[float]  # PDF value of the torsion angle
    min_endpoint_distance_pdf: Optional[float]  # PDF value of the min endpoint distance
    coaxial_probability: Optional[float]  # Probability of stems being coaxial (0-1)


@dataclass
class Structure2D:
    base_pairs: List[BasePair]
    stackings: List[Stacking]
    base_ribose_interactions: List[BaseRibose]
    base_phosphate_interactions: List[BasePhosphate]
    other_interactions: List[OtherInteraction]
    bpseq: BpSeq
    bpseq_index: Dict[int, Residue]
    dot_bracket: MultiStrandDotBracket
    extended_dot_bracket: str
    stems: List[Stem]
    single_strands: List[SingleStrand]
    hairpins: List[Hairpin]
    loops: List[Loop]
    inter_stem_parameters: List[InterStemParameters]



import io
import os

import numpy as np
import pandas as pd
import pytest

from rnapolis.parser_v2 import (
    fit_to_pdb,
    parse_cif_atoms,
    parse_pdb_atoms,
    write_cif,
    write_pdb,
)
from rnapolis.tertiary_v2 import Structure, calculate_torsion_angle


def compare_structures(df1: pd.DataFrame, df2: pd.DataFrame, rtol=1e-5, atol=1e-8):
    """
    Compares two structures derived from atom DataFrames.

    Constructs Structure objects and compares them residue by residue,
    and atom by atom within each residue.

    Parameters:
    -----------
    df1, df2 : pd.DataFrame
        Input DataFrames containing atom data.
    rtol : float, optional
        Relative tolerance for comparing floating-point atom coordinates.
    atol : float, optional
        Absolute tolerance for comparing floating-point atom coordinates.
    """
    struct1 = Structure(df1)
    struct2 = Structure(df2)

    residues1 = struct1.residues
    residues2 = struct2.residues

    assert len(residues1) == len(residues2), (
        f"Different number of residues: {len(residues1)} vs {len(residues2)}"
    )

    # Sort residues for consistent comparison order
    def key_func(r):
        return (
            r.chain_id,
            r.residue_number,
            r.insertion_code or "",
        )

    residues1.sort(key=key_func)
    residues2.sort(key=key_func)

    for i, (res1, res2) in enumerate(zip(residues1, residues2)):
        res1_id = (res1.chain_id, res1.residue_number, res1.insertion_code)
        res2_id = (res2.chain_id, res2.residue_number, res2.insertion_code)

        assert res1_id == res2_id, (
            f"Residue identifier mismatch at index {i}: {res1_id} vs {res2_id}"
        )
        assert res1.residue_name == res2.residue_name, (
            f"Residue name mismatch for {res1_id}: "
            f"{res1.residue_name} vs {res2.residue_name}"
        )

        atoms1 = res1.atoms_list
        atoms2 = res2.atoms_list

        assert len(atoms1) == len(atoms2), (
            f"Different number of atoms in residue {res1_id}: "
            f"{len(atoms1)} vs {len(atoms2)}"
        )

        # Sort atoms by name for consistent comparison
        atoms1.sort(key=lambda a: a.name)
        atoms2.sort(key=lambda a: a.name)

        for j, (atom1, atom2) in enumerate(zip(atoms1, atoms2)):
            assert atom1.name == atom2.name, (
                f"Atom name mismatch in residue {res1_id} at atom index {j}: "
                f"{atom1.name} vs {atom2.name}"
            )
            # Allow element comparison to be case-insensitive if needed, but strict for now
            assert atom1.element == atom2.element, (
                f"Atom element mismatch for atom {atom1.name} in residue {res1_id}: "
                f"{atom1.element} vs {atom2.element}"
            )

            # Compare coordinates with tolerance
            np.testing.assert_allclose(
                atom1.coordinates,
                atom2.coordinates,
                rtol=rtol,
                atol=atol,
                err_msg=(
                    f"Atom coordinate mismatch for atom {atom1.name} in residue {res1_id}"
                ),
            )


@pytest.fixture
def data_dir():
    ("serial",)
    ("name",)
    ("altLoc",)
    ("resName",)
    ("chainID",)
    ("resSeq",)
    ("iCode",)
    ("x",)
    ("y",)
    ("z",)
    ("occupancy",)
    ("tempFactor",)
    ("element",)
    ("charge",)
    """Return the path to the test data directory."""
    return os.path.join(os.path.dirname(__file__))


def test_parse_4qln_formats(data_dir):
    """Test parsing 4qln in both PDB and mmCIF formats and compare residues and torsion angles."""
    # Load PDB and mmCIF files
    pdb_path = os.path.join(data_dir, "4qln.pdb")
    cif_path = os.path.join(data_dir, "4qln.cif")

    # Skip test if files don't exist
    if not (os.path.exists(pdb_path) and os.path.exists(cif_path)):
        pytest.skip(f"Test files not found: {pdb_path} or {cif_path}")

    # Parse both formats
    with open(pdb_path, "r") as pdb_file:
        pdb_atoms = parse_pdb_atoms(pdb_file)

    with open(cif_path, "r") as cif_file:
        cif_atoms = parse_cif_atoms(cif_file)

    # Create structures
    pdb_structure = Structure(pdb_atoms)
    cif_structure = Structure(cif_atoms)

    # Get residues
    pdb_residues = pdb_structure.residues
    cif_residues = cif_structure.residues

    # Basic checks
    assert len(pdb_residues) > 0, "No residues found in PDB file"
    assert len(cif_residues) > 0, "No residues found in mmCIF file"

    # Compare residue counts
    assert len(pdb_residues) == len(cif_residues), (
        f"Different number of residues: PDB={len(pdb_residues)}, mmCIF={len(cif_residues)}"
    )

    # Compare residue identifiers
    pdb_residue_ids = [
        (r.chain_id, r.residue_number, r.insertion_code) for r in pdb_residues
    ]
    cif_residue_ids = [
        (r.chain_id, r.residue_number, r.insertion_code) for r in cif_residues
    ]

    # Sort both lists to ensure consistent ordering
    pdb_residue_ids.sort()
    cif_residue_ids.sort()

    # Check if residue identifiers match
    for i, (pdb_id, cif_id) in enumerate(zip(pdb_residue_ids, cif_residue_ids)):
        assert pdb_id == cif_id, (
            f"Residue mismatch at position {i}: PDB={pdb_id}, mmCIF={cif_id}"
        )

    # Create a mapping from residue ID to residue name for both formats
    pdb_id_to_name = {
        (r.chain_id, r.residue_number, r.insertion_code): r.residue_name
        for r in pdb_residues
    }
    cif_id_to_name = {
        (r.chain_id, r.residue_number, r.insertion_code): r.residue_name
        for r in cif_residues
    }

    # Check if residue names match for each residue ID
    for res_id in pdb_id_to_name:
        assert res_id in cif_id_to_name, f"Residue ID {res_id} not found in mmCIF"
        assert pdb_id_to_name[res_id] == cif_id_to_name[res_id], (
            f"Residue name mismatch for {res_id}: PDB={pdb_id_to_name[res_id]}, mmCIF={cif_id_to_name[res_id]}"
        )

    # Calculate torsion angles for both structures
    pdb_torsion_df = pdb_structure.torsion_angles
    cif_torsion_df = cif_structure.torsion_angles

    # Check if torsion angle DataFrames have the same shape
    assert pdb_torsion_df.shape == cif_torsion_df.shape, (
        f"Different torsion angle DataFrame shapes: PDB={pdb_torsion_df.shape}, mmCIF={cif_torsion_df.shape}"
    )

    # Sort both DataFrames by chain_id, residue_number, and insertion_code for consistent comparison
    pdb_torsion_df = pdb_torsion_df.sort_values(
        by=["chain_id", "residue_number", "insertion_code"]
    ).reset_index(drop=True)

    cif_torsion_df = cif_torsion_df.sort_values(
        by=["chain_id", "residue_number", "insertion_code"]
    ).reset_index(drop=True)

    # Compare residue identifiers in torsion angle DataFrames
    pd.testing.assert_series_equal(
        pdb_torsion_df["chain_id"],
        cif_torsion_df["chain_id"],
        check_names=False,
        check_dtype=False,
    )
    pd.testing.assert_series_equal(
        pdb_torsion_df["residue_number"],
        cif_torsion_df["residue_number"],
        check_names=False,
        check_dtype=False,
    )
    pd.testing.assert_series_equal(
        pdb_torsion_df["residue_name"],
        cif_torsion_df["residue_name"],
        check_names=False,
        check_dtype=False,
    )

    # Compare torsion angle values with a tolerance
    angle_columns = ["alpha", "beta", "gamma", "delta", "epsilon", "zeta", "chi"]
    for col in angle_columns:
        # Skip columns that might not exist in both DataFrames
        if col not in pdb_torsion_df.columns or col not in cif_torsion_df.columns:
            continue

        # Get non-NaN values that exist in both DataFrames
        pdb_values = pdb_torsion_df[col]
        cif_values = cif_torsion_df[col]

        # Check if the same values are NaN in both DataFrames
        assert pdb_values.isna().equals(cif_values.isna()), (
            f"Different NaN patterns in {col} angle"
        )

        # Compare non-NaN values with tolerance
        mask = ~pdb_values.isna()
        if mask.any():
            pdb_non_nan = pdb_values[mask].values
            cif_non_nan = cif_values[mask].values

            # Allow a small tolerance for floating-point differences
            np.testing.assert_allclose(
                pdb_non_nan,
                cif_non_nan,
                rtol=1e-5,
                atol=1e-5,
                err_msg=f"Torsion angle values for {col} don't match between PDB and mmCIF",
            )


def test_torsion_angle_calculation():
    """Test the torsion angle calculation function."""
    # Define four points that form a known torsion angle
    a1 = np.array([1.0, 0.0, 0.0])
    a2 = np.array([0.0, 0.0, 0.0])
    a3 = np.array([0.0, 1.0, 0.0])
    a4 = np.array([0.0, 1.0, 1.0])

    # Calculate the torsion angle

    angle = calculate_torsion_angle(a1, a2, a3, a4)

    # The expected angle is pi/2 radians (90 degrees)
    assert abs(angle - np.pi / 2) < 1e-6, (
        f"Expected angle close to pi/2 radians, got {angle}"
    )

    # Test with collinear points
    a1 = np.array([0.0, 0.0, 0.0])
    a2 = np.array([1.0, 0.0, 0.0])
    a3 = np.array([2.0, 0.0, 0.0])
    a4 = np.array([3.0, 0.0, 0.0])

    angle = calculate_torsion_angle(a1, a2, a3, a4)
    assert np.isnan(angle), f"Expected NaN for collinear points, got {angle}"


def test_connected_residues_and_torsion_angles(data_dir):
    """Test finding connected residues and calculating torsion angles."""
    # Load PDB file
    pdb_path = os.path.join(data_dir, "4qln.pdb")

    # Skip test if file doesn't exist
    if not os.path.exists(pdb_path):
        pytest.skip(f"Test file not found: {pdb_path}")

    # Parse PDB file
    with open(pdb_path, "r") as pdb_file:
        pdb_atoms = parse_pdb_atoms(pdb_file)

    # Create structure
    structure = Structure(pdb_atoms)

    # Find connected residues
    segments = structure.connected_residues

    # Check that we found at least one segment
    assert len(segments) > 0, "No connected residue segments found"

    # Check that each segment has at least 2 residues
    for segment in segments:
        assert len(segment) >= 2, f"Segment has fewer than 2 residues: {segment}"

    # Calculate torsion angles
    torsion_df = structure.torsion_angles

    # Check that the DataFrame has the expected columns
    expected_columns = [
        "chain_id",
        "residue_number",
        "insertion_code",
        "residue_name",
        "alpha",
        "beta",
        "gamma",
        "delta",
        "epsilon",
        "zeta",
        "chi",
    ]
    for col in expected_columns:
        assert col in torsion_df.columns, (
            f"Expected column {col} not found in torsion angles DataFrame"
        )

    # Check that we have some torsion angle values
    assert len(torsion_df) > 0, "No torsion angles calculated"

    # Check that at least some angles are not null
    for angle in ["alpha", "beta", "gamma", "delta", "epsilon", "zeta", "chi"]:
        assert torsion_df[angle].notna().any(), f"No valid {angle} angles calculated"


def test_pdb_cif_pdb_roundtrip(data_dir):
    """Test PDB -> CIF -> PDB conversion preserves essential data."""
    pdb_path = os.path.join(data_dir, "1ATO.pdb")
    if not os.path.exists(pdb_path):
        pytest.skip(f"Test file not found: {pdb_path}")

    # 1. Parse Original PDB
    with open(pdb_path, "r") as f:
        df_orig_pdb = parse_pdb_atoms(f)
    assert not df_orig_pdb.empty, "Original PDB parsing failed"

    # 2. Write CIF
    cif_str = write_cif(df_orig_pdb)
    assert cif_str is not None and "atom_site" in cif_str, "Writing CIF failed"

    # 3. Parse Intermediate CIF
    df_intermediate_cif = parse_cif_atoms(io.StringIO(cif_str))
    assert not df_intermediate_cif.empty, "Parsing intermediate CIF failed"

    # 4. Write Final PDB
    pdb_final_str = write_pdb(df_intermediate_cif)
    assert pdb_final_str is not None and "ATOM" in pdb_final_str, (
        "Writing final PDB failed"
    )

    # 5. Parse Final PDB
    df_final_pdb = parse_pdb_atoms(io.StringIO(pdb_final_str))
    assert not df_final_pdb.empty, "Parsing final PDB failed"

    # 6. Compare Original PDB Structure with Final PDB Structure
    compare_structures(df_orig_pdb, df_final_pdb)


def test_cif_pdb_cif_roundtrip(data_dir):
    """Test CIF -> PDB -> CIF conversion preserves essential data."""
    cif_path = os.path.join(data_dir, "1ehz-assembly-1.cif")
    if not os.path.exists(cif_path):
        pytest.skip(f"Test file not found: {cif_path}")

    # 1. Parse Original CIF
    with open(cif_path, "r") as f:
        df_orig_cif = parse_cif_atoms(f)
    assert not df_orig_cif.empty, "Original CIF parsing failed"

    # 2. Fit to PDB (if necessary) and Write PDB
    try:
        # fit_to_pdb returns original df if it fits, otherwise a modified copy
        df_to_write_pdb = fit_to_pdb(df_orig_cif)
        pdb_str = write_pdb(df_to_write_pdb)
        assert pdb_str is not None and "ATOM" in pdb_str, "Writing PDB failed"
    except ValueError as e:
        pytest.skip(f"Cannot fit {cif_path} to PDB, skipping roundtrip test: {e}")

    # 3. Parse Intermediate PDB
    df_from_pdb = parse_pdb_atoms(io.StringIO(pdb_str))
    assert not df_from_pdb.empty, "Parsing intermediate PDB failed"

    # 4. Write Final CIF
    cif_final_str = write_cif(df_from_pdb)
    assert cif_final_str is not None and "atom_site" in cif_final_str, (
        "Writing final CIF failed"
    )

    # 5. Parse Final CIF
    df_final_cif = parse_cif_atoms(io.StringIO(cif_final_str))
    assert not df_final_cif.empty, "Parsing final CIF failed"

    # 6. Compare Original CIF Structure with Final CIF Structure
    # Note: This comparison might fail if fit_to_pdb modified the structure
    # (e.g., renumbered chains/residues). The comparison checks if the
    # *essential structural information* (residue names, atom names, coords)
    # is preserved through the PDB intermediate representation, assuming
    # the PDB format could represent the original structure or its fitted version.
    compare_structures(df_to_write_pdb, df_final_cif)


def test_groupby_order(data_dir):
    cif_path = os.path.join(data_dir, "1ehz-assembly-1.cif")
    if not os.path.exists(cif_path):
        pytest.skip(f"Test file not found: {cif_path}")

    # 1. Parse Original CIF
    with open(cif_path, "r") as f:
        df_atoms = parse_cif_atoms(f)
    assert not df_atoms.empty, "Original CIF parsing failed"

    # 2. Group by chain_id and residue_number
    structure = Structure(df_atoms)  # Ensure Structure can be created
    residues = structure.residues  # Access residues to trigger any necessary processing
    assert str(residues[0]) == "A.G1"
    assert str(residues[1]) == "A.C2"


def test_is_nucleotide(data_dir):
    cif_path = os.path.join(data_dir, "1ehz-assembly-1.cif")
    if not os.path.exists(cif_path):
        pytest.skip(f"Test file not found: {cif_path}")

    with open(cif_path, "r") as f:
        df_atoms = parse_cif_atoms(f)
    assert not df_atoms.empty, "Original CIF parsing failed"

    structure = Structure(df_atoms)
    residues = [residue for residue in structure.residues if residue.is_nucleotide]
    assert len(residues) == 76



import re
import tempfile

from rnapolis.molecule_filter import filter_by_chains, filter_by_poly_types
from rnapolis.parser import parse_cif


def test_filter_by_poly_types():
    with open("tests/1a9n.cif") as f:
        content = f.read()

    filtered = filter_by_poly_types(content, ["polyribonucleotide"], ["chem_comp"])
    assert re.search(r"^_entity\.id", filtered, re.MULTILINE) is not None
    assert re.search(r"^_chem_comp\.id", filtered, re.MULTILINE) is not None

    with tempfile.NamedTemporaryFile("rt+") as f:
        f.write(filtered)
        f.seek(0)
        atoms, _, _, _ = parse_cif(f)

    chains = set([atom.label.chain for atom in atoms if atom.label])
    assert chains == {"A", "B"}


def test_filter_by_chains():
    with open("tests/1a9n.cif") as f:
        content = f.read()

    filtered = filter_by_chains(content, ["A", "C"], ["chem_comp"])
    assert re.search(r"^_entity\.id", filtered, re.MULTILINE) is not None
    assert re.search(r"^_chem_comp\.id", filtered, re.MULTILINE) is not None

    with tempfile.NamedTemporaryFile("rt+") as f:
        f.write(filtered)
        f.seek(0)
        atoms, _, _, _ = parse_cif(f)

    chains = set([atom.label.chain for atom in atoms if atom.label])
    assert chains >= {"A", "C"}



from collections import Counter

from rnapolis.annotator import extract_base_interactions
from rnapolis.parser import read_3d_structure


def test_1ehz():
    """
    Make sure there are no duplicates on any list
    """
    with open("tests/1ehz-assembly-1.cif") as f:
        structure3d = read_3d_structure(f, 1)
    base_interactions = extract_base_interactions(structure3d, 1)

    interactions = [
        base_interactions.base_pairs,
        base_interactions.stackings,
        base_interactions.base_ribose_interactions,
        base_interactions.base_phosphate_interactions,
        base_interactions.other_interactions,
    ]
    labels = ["base pairs", "stackings", "base-ribose", "base-phosphate", "other"]
    for i in range(len(interactions)):
        counter = Counter(interactions[i])
        for element, count in counter.most_common():
            assert count == 1, f"Interaction {element} occurs {count} times"

        simple_collection = [
            (bp.nt1.full_name, bp.nt2.full_name) for bp in interactions[i]
        ]
        counter = Counter(simple_collection)
        for element, count in counter.most_common():
            if count != 1:
                duplicates = [
                    bp
                    for bp in interactions[i]
                    if (bp.nt1.full_name, bp.nt2.full_name) == element
                ]
                assert False, (
                    f"Interaction {element} occurs {count} times among {labels[i]} type: {duplicates}"
                )


def test_8btk():
    with open("tests/8btk_B7.cif") as f:
        structure3d = read_3d_structure(f, 1)
    base_interactions = extract_base_interactions(structure3d, 1)
    assert structure3d.extract_secondary_structure(base_interactions) is not None


def test_488d():
    """
    There are clashing residues 151 in chains B and D. The clash is caused by occupancy factors less than 1.
    """
    with open("tests/488d.pdb") as f:
        structure3d = read_3d_structure(f)

    base_interactions = extract_base_interactions(structure3d)
    assert base_interactions is not None



import math

from rnapolis.parser import read_3d_structure
from rnapolis.tertiary import Atom, torsion_angle


def test_torsion_angle():
    a1 = Atom(None, None, None, 1, "P", 50.63, 49.73, 50.57, None)
    a2 = Atom(None, None, None, 1, "O5'", 50.16, 49.14, 52.02, None)
    a3 = Atom(None, None, None, 1, "C5'", 50.22, 49.95, 53.21, None)
    a4 = Atom(None, None, None, 1, "C4'", 50.97, 49.23, 54.31, None)
    assert math.isclose(
        math.degrees(torsion_angle(a1, a2, a3, a4)), -127.83976634524326
    )


def test_nucleobase_atoms():
    with open("tests/1E7K_1_C.cif") as f:
        structure3d = read_3d_structure(f)
    for residue in structure3d.residues:
        assert residue.has_all_nucleobase_heavy_atoms

    with open("tests/1E7K_1_C_modified.cif") as f:
        structure3d = read_3d_structure(f)
    assert all(
        not residue.has_all_nucleobase_heavy_atoms for residue in structure3d.residues
    )



import gzip

from rnapolis.parser import read_3d_structure


def test_nucleic_acid_only():
    with open("tests/184D.cif") as f:
        structure3d = read_3d_structure(f, 1, nucleic_acid_only=False)
    assert len(structure3d.residues) == 114

    with open("tests/184D.cif") as f:
        structure3d = read_3d_structure(f, 1, nucleic_acid_only=True)
    assert len(structure3d.residues) == 14


def test_1ato():
    with open("tests/1ATO.pdb") as f:
        structure3d = read_3d_structure(f)
    sequence = "".join([residue.one_letter_name for residue in structure3d.residues])
    assert sequence == "GGCACCUCCUCGCGGUGCC"


def test_4qln_no_duplicate_atoms():
    for ext in (".pdb", ".cif"):
        with open(f"tests/4qln{ext}") as f:
            structure3d = read_3d_structure(f)

        chain_a = [r for r in structure3d.residues if r.auth.chain == "A"]
        residues_to_check = [r for r in chain_a if r.auth.number in (18, 19, 20)]

        for residue in residues_to_check:
            atom_names = [atom.name for atom in residue.atoms]
            assert len(atom_names) == len(set(atom_names)), (
                f"Duplicate atoms found in residue {residue.auth}"
            )


def test_1gid():
    expected_sequence = "GAAUUGCGGGAAAGGGGUCAACAGCCGUUCAGUACCAAGUCUCAGGGGAAACUUUGAGAUGGCCUUGCAAAGGGUAUGGUAAUAAGCUGACGGACAUGGUCCUAACCACGCAGCCAAGUCCUAAGUCAACAGAUCUUCUGUUGAUAUGGAUGCAGUUC"

    with gzip.open("tests/1gid.cif.gz", "rt") as f:
        structure3d = read_3d_structure(f, nucleic_acid_only=True)

    residues_a = [r for r in structure3d.residues if r.auth.chain == "A"]
    residues_b = [r for r in structure3d.residues if r.auth.chain == "B"]
    assert len(residues_a) == len(expected_sequence)
    assert len(residues_b) == len(expected_sequence)

    actual_sequence_a = "".join([residue.one_letter_name for residue in residues_a])
    actual_sequence_b = "".join([residue.one_letter_name for residue in residues_b])
    assert actual_sequence_a == expected_sequence
    assert actual_sequence_b == expected_sequence



import argparse
import csv
import logging
import math
import os
from enum import Enum
from functools import cached_property
from typing import List, Optional

import numpy as np
from rnapolis.metareader import read_metadata
from rnapolis.parser import read_3d_structure
from rnapolis.tertiary import Atom, Residue3D
from scipy.spatial import KDTree

CARBON_RADIUS = 0.6
NITROGEN_RADIUS = 0.54
OXYGEN_RADIUS = 0.53
PHOSPHORUS_RADIUS = 0.94

logging.basicConfig(level=os.getenv("LOGLEVEL", "INFO").upper())


class AtomType(Enum):
    C = "C"
    N = "N"
    O = "O"
    P = "P"

    @cached_property
    def radius(self) -> float:
        if self.value == "C":
            return CARBON_RADIUS
        elif self.value == "N":
            return NITROGEN_RADIUS
        elif self.value == "O":
            return OXYGEN_RADIUS
        elif self.value == "P":
            return PHOSPHORUS_RADIUS
        raise RuntimeError(f"Unknown atom type: {self}")

    def matches(self, atom: Atom):
        return atom.name.strip().startswith(self.value)


def find_clashes(
    residues: List[Residue3D],
    ignore_occupancy: bool,
    ignore_autoclashes: bool,
    nucleic_acid_only: bool,
    require_same_atom_name: bool,
    enable_molprobity_mode: bool,
):
    reference_residues = []
    reference_atoms = []
    coordinates = []

    for residue in residues:
        if (
            nucleic_acid_only is True and residue.is_nucleotide
        ) or nucleic_acid_only is False:
            for atom in residue.atoms:
                if any([atom_type.matches(atom) for atom_type in AtomType]):
                    reference_residues.append(residue)
                    reference_atoms.append(atom)
                    coordinates.append(atom.coordinates)

    if len(coordinates) < 2:
        return []

    kdtree = KDTree(coordinates)
    result = []
    max_radius = max([atom_type.radius for atom_type in AtomType])
    molprobity_factor = 0.5 if enable_molprobity_mode is True else 0.0

    for i, j in kdtree.query_pairs(2.0 * max_radius + molprobity_factor):
        ai: Atom = reference_atoms[i]
        aj: Atom = reference_atoms[j]
        ri, rj = reference_residues[i], reference_residues[j]

        if ignore_autoclashes is True and ri == rj:
            continue
        if require_same_atom_name is True and ai.name != aj.name:
            continue

        distance = np.linalg.norm(ai.coordinates - aj.coordinates)
        sum_vdw_radii = AtomType[ai.name[0]].radius + AtomType[aj.name[0]].radius
        if distance > sum_vdw_radii + molprobity_factor:
            continue

        sum_occupancies = (ai.occupancy or 1.0) + (aj.occupancy or 1.0)
        if ignore_occupancy is True or math.isclose(sum_occupancies, 1.0):
            result.append(((ri, ai), (rj, aj), sum_occupancies))

    return result


def classify_clash(atom_i: Atom, atom_j: Atom, occupancy: float) -> Optional[str]:
    if atom_i.name == "O3'" and atom_j.name in (
        "OP1",
        "OP2",
        "OP3",
        "O1P",
        "O2P",
        "O3P",
    ):
        return "O3'"
    return None


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help="Path to PDB or mmCIF file")
    parser.add_argument(
        "--ignore-occupancy",
        help="By default clashes are reported if atoms' occupancies are not equal to 1.0, but you can ignore this check. If you ignore this check, any pair of atoms too close to each other will be reported regardless of their occupancy",
        action="store_true",
    )
    parser.add_argument(
        "--nucleic-acid-only",
        help="By default all kind of clashes will be found, but you can focus only on nucleic acids chains",
        action="store_true",
    )
    parser.add_argument(
        "--ignore-autoclashes",
        help="By default clashes will be reported even in scope of the same residue, but you can disable this behaviour",
        action="store_true",
    )
    parser.add_argument(
        "--require-same-atom-name",
        help="By default any two clashing atoms are reported (e.g. P vs OP1), but when this is set, the program will report only clashes of the same atom name (e.g. OP1 vs OP1)",
        action="store_true",
    )
    parser.add_argument(
        "--enable-molprobity-mode",
        help="By default this tool will report any *strong* clash, i.e., when two atoms are closer than their sum of vdW radii; when this option is set, additional 0.5A is added as in MolProbity, so that more clashes are detected, but some of them might be *weak*",
        action="store_true",
    )
    parser.add_argument("--csv", help="Store result in CSV format")
    args = parser.parse_args()

    with open(args.input) as f:
        structure3d = read_3d_structure(f, 1)

    clashing_chains = {}
    max_occupancy_residues = {}
    max_occupancy_chains = {}

    clashes = find_clashes(
        structure3d.residues,
        args.ignore_occupancy,
        args.ignore_autoclashes,
        args.nucleic_acid_only,
        args.require_same_atom_name,
        args.enable_molprobity_mode,
    )

    if clashes:
        for pi, pj, occupancy in clashes:
            ri, ai = pi
            rj, aj = pj

            chain_key = (ri.chain, rj.chain)
            residue_key = (ri, rj)
            if chain_key not in clashing_chains:
                clashing_chains[chain_key] = {}
            if residue_key not in clashing_chains[chain_key]:
                clashing_chains[chain_key][residue_key] = set()
            clashing_chains[chain_key][residue_key].add((ai, aj, occupancy))

            max_occupancy_residues[(ri, rj)] = max(
                [max_occupancy_residues.get((ri, rj), 0.0), occupancy]
            )
            max_occupancy_chains[(ri.chain, rj.chain)] = max(
                [max_occupancy_residues.get((ri.chain, rj.chain), 0.0), occupancy]
            )

    if clashing_chains:
        for ci, cj in sorted(clashing_chains):
            if ci == cj:
                print(
                    f"Clashes found in chain {ci} with maximum occupancy sum equal to {max_occupancy_chains[(ci, cj)]}"
                )
            else:
                print(
                    f"Clashes found between chains {ci} and {cj} with maximum occupancy sum equal to {max_occupancy_chains[(ci, cj)]}"
                )
            for ri, rj in clashing_chains[(ci, cj)]:
                if ri == rj:
                    print(
                        f"    Clashes found in residue {ri} with maximum occupancy sum equal to {max_occupancy_residues[(ri, rj)]}"
                    )
                else:
                    print(
                        f"    Clashes found between residues {ri} and {rj} with maximum occupancy sum equal to {max_occupancy_residues[(ri, rj)]}"
                    )
                for ai, aj, occupancy in sorted(clashing_chains[(ci, cj)][(ri, rj)]):
                    print(
                        f"        Clashes found between atoms {ai.name} and {aj.name} with occupancy sum of {occupancy}"
                    )

        if args.csv:
            metadata = read_metadata(args.input, ["exptl", "refine"])

            with open(args.csv, "w") as f:
                writer = csv.writer(f)
                writer.writerow(
                    [
                        "Filename",
                        "Experimental method",
                        "Resolution",
                        "Atom 1",
                        "Atom 2",
                        "Occupancy sum",
                        "Classification",
                    ]
                )

                for ci, cj in sorted(clashing_chains):
                    for ri, rj in clashing_chains[(ci, cj)]:
                        for ai, aj, occupancy in sorted(
                            clashing_chains[(ci, cj)][(ri, rj)]
                        ):
                            writer.writerow(
                                [
                                    f"{os.path.splitext(os.path.basename(args.input))[0]}",
                                    metadata["exptl"][0]["method"],
                                    metadata["refine"][0]["ls_d_res_high"],
                                    f"{ri} {ai.name}",
                                    f"{rj} {aj.name}",
                                    occupancy,
                                    classify_clash(ai, aj, occupancy),
                                ]
                            )


if __name__ == "__main__":
    main()



#! /usr/bin/env python
import argparse
import string
import tempfile
from typing import Dict, Tuple

from mmcif.io.IoAdapterPy import IoAdapterPy
from mmcif.io.PdbxReader import DataCategory


def copy_from_to(
    file_content: str,
    category: str = "atom_site",
    copy_from: str = "label_asym_id",
    copy_to: str = "auth_asym_id",
) -> str:
    adapter = IoAdapterPy()

    with tempfile.NamedTemporaryFile(mode="wt") as f:
        f.write(file_content)
        f.seek(0)
        data = adapter.readFile(f.name)

    if len(data) == 0 or category not in data[0].getObjNameList():
        return file_content

    category_obj = data[0].getObj(category)
    attributes = category_obj.getAttributeList()

    if copy_from not in attributes:
        return file_content

    transformed = []

    if copy_to not in attributes:
        attributes.append(copy_to)

    for row in category_obj.getRowList():
        i = attributes.index(copy_from)
        j = attributes.index(copy_to)
        if j >= len(row):
            row.append(row[i])
        else:
            row[j] = row[i]
        transformed.append(row)

    data[0].replace(DataCategory(category_obj, attributes, transformed))

    with tempfile.NamedTemporaryFile(mode="rt+") as f:
        adapter.writeFile(f.name, data)
        f.seek(0)
        return f.read()


def replace_value(
    file_content: str,
    category: str = "atom_site",
    column: str = "auth_asym_id",
    values: str = "".join([c for c in string.printable if c not in string.whitespace]),
) -> Tuple[str, Dict]:
    adapter = IoAdapterPy()
    with tempfile.NamedTemporaryFile(mode="wt") as f:
        f.write(file_content)
        f.seek(0)
        data = adapter.readFile(f.name)

    if len(data) == 0 or category not in data[0].getObjNameList():
        return file_content, {}

    category_obj = data[0].getObj(category)
    attributes = category_obj.getAttributeList()

    if column not in attributes:
        return file_content, {}

    transformed = []
    mapping = {}

    for row in category_obj.getRowList():
        i = attributes.index(column)

        if row[i] not in mapping:
            mapping[row[i]] = values[len(mapping)]

        row[i] = mapping[row[i]]
        transformed.append(row)

    data[0].replace(DataCategory(category_obj, attributes, transformed))

    with tempfile.NamedTemporaryFile(mode="rt+") as f:
        adapter.writeFile(f.name, data)
        f.seek(0)
        return f.read(), mapping


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help="path to input mmCIF file")
    parser.add_argument("output", help="path to output mmCIF file")
    parser.add_argument(
        "--category", help="name of the category to work on, e.g., atom_site"
    )
    parser.add_argument(
        "--copy-from",
        help="name of a data item to copy from, e.g., label_asym_id (exclusive with --replace)",
    )
    parser.add_argument(
        "--copy-to",
        help="name of a data item to copy to, e.g., auth_asym_id (exclusive with --replace)",
    )
    parser.add_argument(
        "--replace",
        help="name of a data item to replace values, e.g., auth_asym_id (exclusive with --copy-from and --copy-to)",
    )
    parser.add_argument(
        "--values",
        help="values to replace with, e.g., ABCDEFGHIJKLMNOPQRSTUVWXYZ (exclusive with --copy-from and --copy-to)",
    )
    args = parser.parse_args()

    if args.copy_from and args.copy_to:
        output = copy_from_to(args.input, args.category, args.copy_from, args.copy_to)
    elif args.replace and args.values:
        output = replace_value(args.input, args.category, args.replace, args.values)
    else:
        parser.print_help()
        return

    with open(args.output, "w") as f:
        f.write(output)


if __name__ == "__main__":
    main()



import os

import pytest
from rnapolis.rfam_folder import generate_consensus_secondary_structure, parse_fasta

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Test doesn't work in Github Actions.")
def test_GK000002():
    fasta = parse_fasta("tests/GK000002.2-66269475-66272524.fa")
    generate_consensus_secondary_structure(fasta[0], "RF02540", False)


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Test doesn't work in Github Actions.")
def test_LJAJ01000009():
    fasta = parse_fasta("tests/LJAJ01000009.1-42413-42384.fa")
    result = generate_consensus_secondary_structure(fasta[0], "RF01315", False)
    sequence = result[0].split("\n")[1]
    assert len(sequence) == len(fasta[0].sequence)


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Test doesn't work in Github Actions.")
def test_RF03339():
    fastas = parse_fasta("tests/RF03339.fa")
    for fasta in fastas:
        result = generate_consensus_secondary_structure(fasta, "RF03339", False)
        sequence = result[0].split("\n")[1]
        assert len(sequence) == len(fasta.sequence)



from rnapolis.annotator import extract_base_interactions
from rnapolis.common import ResidueAuth, ResidueLabel
from rnapolis.parser import read_3d_structure
from rnapolis.tertiary import Mapping2D3D


# 2HY9 has a canonical quadruplex
def test_2HY9():
    with open("tests/2HY9.cif") as f:
        structure3d = read_3d_structure(f, 1)
    base_interactions = extract_base_interactions(structure3d)
    mapping = Mapping2D3D(
        structure3d, base_interactions.base_pairs, base_interactions.stackings, True
    )

    # tract 1
    g4 = structure3d.find_residue(ResidueLabel("A", 4, "DG"), None)
    g5 = structure3d.find_residue(ResidueLabel("A", 5, "DG"), None)
    g6 = structure3d.find_residue(ResidueLabel("A", 6, "DG"), None)

    # tract 2
    g10 = structure3d.find_residue(ResidueLabel("A", 10, "DG"), None)
    g11 = structure3d.find_residue(ResidueLabel("A", 11, "DG"), None)
    g12 = structure3d.find_residue(ResidueLabel("A", 12, "DG"), None)

    # tract 3
    g16 = structure3d.find_residue(ResidueLabel("A", 16, "DG"), None)
    g17 = structure3d.find_residue(ResidueLabel("A", 17, "DG"), None)
    g18 = structure3d.find_residue(ResidueLabel("A", 18, "DG"), None)

    # tract 4
    g22 = structure3d.find_residue(ResidueLabel("A", 22, "DG"), None)
    g23 = structure3d.find_residue(ResidueLabel("A", 23, "DG"), None)
    g24 = structure3d.find_residue(ResidueLabel("A", 24, "DG"), None)

    assert all(
        nt is not None
        for nt in [g4, g5, g6, g10, g11, g12, g16, g17, g18, g22, g23, g24]
    )

    # tetrad 1
    assert {g10, g22}.issubset(mapping.base_pair_graph[g4])  # type: ignore
    assert {g10, g22}.issubset(mapping.base_pair_graph[g18])  # type: ignore
    assert {g4, g18}.issubset(mapping.base_pair_graph[g10])  # type: ignore
    assert {g4, g18}.issubset(mapping.base_pair_graph[g22])  # type: ignore

    # tetrad 2
    assert {g11, g23}.issubset(mapping.base_pair_graph[g5])  # type: ignore
    assert {g11, g23}.issubset(mapping.base_pair_graph[g17])  # type: ignore
    assert {g5, g17}.issubset(mapping.base_pair_graph[g11])  # type: ignore
    assert {g5, g17}.issubset(mapping.base_pair_graph[g23])  # type: ignore

    # tetrad 3
    assert {g12, g24}.issubset(mapping.base_pair_graph[g6])  # type: ignore
    assert {g12, g24}.issubset(mapping.base_pair_graph[g16])  # type: ignore
    assert {g6, g16}.issubset(mapping.base_pair_graph[g12])  # type: ignore
    assert {g6, g16}.issubset(mapping.base_pair_graph[g24])  # type: ignore


# 6RS3 has a quadruplex made of modified residues (GF2)
def test_6RS3():
    with open("tests/6RS3.cif") as f:
        structure3d = read_3d_structure(f, 1)
    base_interactions = extract_base_interactions(structure3d)
    mapping = Mapping2D3D(
        structure3d, base_interactions.base_pairs, base_interactions.stackings, True
    )

    g1 = structure3d.find_residue(ResidueLabel("A", 1, "DG"), None)
    g2 = structure3d.find_residue(ResidueLabel("A", 2, "DG"), None)
    g6 = structure3d.find_residue(ResidueLabel("A", 6, "DG"), None)
    g7 = structure3d.find_residue(ResidueLabel("A", 7, "DG"), None)
    g8 = structure3d.find_residue(ResidueLabel("A", 8, "DG"), None)
    g14 = structure3d.find_residue(ResidueLabel("A", 14, "GF2"), None)
    g15 = structure3d.find_residue(ResidueLabel("A", 15, "GF2"), None)
    g16 = structure3d.find_residue(ResidueLabel("A", 16, "DG"), None)
    g17 = structure3d.find_residue(ResidueLabel("A", 17, "DG"), None)
    g20 = structure3d.find_residue(ResidueLabel("A", 20, "DG"), None)
    g21 = structure3d.find_residue(ResidueLabel("A", 21, "DG"), None)
    g22 = structure3d.find_residue(ResidueLabel("A", 22, "DG"), None)

    assert all(
        nt is not None for nt in [g1, g2, g6, g7, g8, g14, g15, g16, g17, g20, g21, g22]
    )

    # tetrad 1
    assert {g2, g20}.issubset(mapping.base_pair_graph[g6])  # type: ignore
    assert {g2, g20}.issubset(mapping.base_pair_graph[g15])  # type: ignore
    assert {g6, g15}.issubset(mapping.base_pair_graph[g2])  # type: ignore
    assert {g6, g15}.issubset(mapping.base_pair_graph[g20])  # type: ignore

    # tetrad 2
    assert {g1, g21}.issubset(mapping.base_pair_graph[g7])  # type: ignore
    assert {g1, g21}.issubset(mapping.base_pair_graph[g16])  # type: ignore
    assert {g7, g16}.issubset(mapping.base_pair_graph[g1])  # type: ignore
    assert {g7, g16}.issubset(mapping.base_pair_graph[g21])  # type: ignore

    # tetrad 3
    assert {g14, g22}.issubset(mapping.base_pair_graph[g8])  # type: ignore
    assert {g14, g22}.issubset(mapping.base_pair_graph[g17])  # type: ignore
    assert {g8, g17}.issubset(mapping.base_pair_graph[g14])  # type: ignore
    assert {g8, g17}.issubset(mapping.base_pair_graph[g22])  # type: ignore


# 1JJP has an inter-locked quadruplex made of two chains
def test_1JJP():
    with open("tests/1JJP.cif") as f:
        structure3d = read_3d_structure(f, 1)
    base_interactions = extract_base_interactions(structure3d)
    mapping = Mapping2D3D(
        structure3d, base_interactions.base_pairs, base_interactions.stackings, True
    )

    ag1 = structure3d.find_residue(ResidueLabel("A", 1, "DG"), None)
    ag2 = structure3d.find_residue(ResidueLabel("A", 2, "DG"), None)
    ag3 = structure3d.find_residue(ResidueLabel("A", 3, "DG"), None)
    ag5 = structure3d.find_residue(ResidueLabel("A", 5, "DG"), None)
    ag6 = structure3d.find_residue(ResidueLabel("A", 6, "DG"), None)
    ag10 = structure3d.find_residue(ResidueLabel("A", 10, "DG"), None)
    ag11 = structure3d.find_residue(ResidueLabel("A", 11, "DG"), None)
    ag12 = structure3d.find_residue(ResidueLabel("A", 12, "DG"), None)

    bg1 = structure3d.find_residue(ResidueLabel("B", 1, "DG"), None)
    bg2 = structure3d.find_residue(ResidueLabel("B", 2, "DG"), None)
    bg3 = structure3d.find_residue(ResidueLabel("B", 3, "DG"), None)
    bg5 = structure3d.find_residue(ResidueLabel("B", 5, "DG"), None)
    bg6 = structure3d.find_residue(ResidueLabel("B", 6, "DG"), None)
    bg10 = structure3d.find_residue(ResidueLabel("B", 10, "DG"), None)
    bg11 = structure3d.find_residue(ResidueLabel("B", 11, "DG"), None)
    bg12 = structure3d.find_residue(ResidueLabel("B", 12, "DG"), None)

    assert all(
        nt is not None
        for nt in [
            ag1,
            ag2,
            ag3,
            ag5,
            ag6,
            ag10,
            ag11,
            ag12,
            bg1,
            bg2,
            bg3,
            bg5,
            bg6,
            bg10,
            bg11,
            bg12,
        ]
    )

    # tetrad 1
    assert {ag6, ag12}.issubset(mapping.base_pair_graph[ag3])  # type: ignore
    assert {ag6, ag12}.issubset(mapping.base_pair_graph[ag10])  # type: ignore
    assert {ag3, ag10}.issubset(mapping.base_pair_graph[ag6])  # type: ignore
    assert {ag3, ag10}.issubset(mapping.base_pair_graph[ag12])  # type: ignore

    # tetrad 2
    assert {ag5, ag11}.issubset(mapping.base_pair_graph[bg1])  # type: ignore
    assert {ag5, ag11}.issubset(mapping.base_pair_graph[ag2])  # type: ignore
    assert {bg1, ag2}.issubset(mapping.base_pair_graph[ag5])  # type: ignore
    assert {bg1, ag2}.issubset(mapping.base_pair_graph[ag11])  # type: ignore

    # tetrad 3
    assert {bg5, bg11}.issubset(mapping.base_pair_graph[ag1])  # type: ignore
    assert {bg5, bg11}.issubset(mapping.base_pair_graph[bg2])  # type: ignore
    assert {ag1, bg2}.issubset(mapping.base_pair_graph[bg5])  # type: ignore
    assert {ag1, bg2}.issubset(mapping.base_pair_graph[bg11])  # type: ignore

    # tetrad 4
    assert {bg6, bg12}.issubset(mapping.base_pair_graph[bg3])  # type: ignore
    assert {bg6, bg12}.issubset(mapping.base_pair_graph[bg10])  # type: ignore
    assert {bg3, bg10}.issubset(mapping.base_pair_graph[bg6])  # type: ignore
    assert {bg3, bg10}.issubset(mapping.base_pair_graph[bg12])  # type: ignore


# 6FC9 has a quadruplex and a duplex
def test_6FC9():
    with open("tests/6FC9.cif") as f:
        structure3d = read_3d_structure(f, 1)
    base_interactions = extract_base_interactions(structure3d)
    mapping = Mapping2D3D(
        structure3d, base_interactions.base_pairs, base_interactions.stackings, True
    )

    g1 = structure3d.find_residue(ResidueLabel("A", 1, "DG"), None)
    g2 = structure3d.find_residue(ResidueLabel("A", 2, "DG"), None)
    g5 = structure3d.find_residue(ResidueLabel("A", 5, "DG"), None)
    g6 = structure3d.find_residue(ResidueLabel("A", 6, "DG"), None)
    c7 = structure3d.find_residue(ResidueLabel("A", 7, "DC"), None)
    g8 = structure3d.find_residue(ResidueLabel("A", 8, "DG"), None)
    c9 = structure3d.find_residue(ResidueLabel("A", 9, "DC"), None)
    g10 = structure3d.find_residue(ResidueLabel("A", 10, "DG"), None)
    a11 = structure3d.find_residue(ResidueLabel("A", 11, "DA"), None)
    a12 = structure3d.find_residue(ResidueLabel("A", 12, "DA"), None)
    t16 = structure3d.find_residue(ResidueLabel("A", 16, "DT"), None)
    t17 = structure3d.find_residue(ResidueLabel("A", 17, "DT"), None)
    c18 = structure3d.find_residue(ResidueLabel("A", 18, "DC"), None)
    g19 = structure3d.find_residue(ResidueLabel("A", 19, "DG"), None)
    c20 = structure3d.find_residue(ResidueLabel("A", 20, "DC"), None)
    g21 = structure3d.find_residue(ResidueLabel("A", 21, "DG"), None)
    g22 = structure3d.find_residue(ResidueLabel("A", 22, "DG"), None)
    g23 = structure3d.find_residue(ResidueLabel("A", 23, "DG"), None)
    g26 = structure3d.find_residue(ResidueLabel("A", 26, "DG"), None)
    g27 = structure3d.find_residue(ResidueLabel("A", 27, "DG"), None)

    assert all(
        nt is not None
        for nt in [
            g1,
            g2,
            g5,
            g6,
            c7,
            g8,
            c9,
            g10,
            a11,
            a12,
            t16,
            t17,
            c18,
            g19,
            c20,
            g21,
            g22,
            g23,
            g26,
            g27,
        ]
    )

    # tetrad 1
    assert {g6, g27}.issubset(mapping.base_pair_graph[g1])  # type: ignore
    assert {g6, g27}.issubset(mapping.base_pair_graph[g22])  # type: ignore
    assert {g1, g22}.issubset(mapping.base_pair_graph[g6])  # type: ignore
    assert {g1, g22}.issubset(mapping.base_pair_graph[g27])  # type: ignore

    # tetrad 2
    assert {g5, g26}.issubset(mapping.base_pair_graph[g2])  # type: ignore
    assert {g5, g26}.issubset(mapping.base_pair_graph[g23])  # type: ignore
    assert {g2, g23}.issubset(mapping.base_pair_graph[g5])  # type: ignore
    assert {g2, g23}.issubset(mapping.base_pair_graph[g26])  # type: ignore

    assert mapping.base_pair_dict[(c7, g21)].is_canonical  # type: ignore
    assert mapping.base_pair_dict[(g8, c20)].is_canonical  # type: ignore
    assert mapping.base_pair_dict[(c9, g19)].is_canonical  # type: ignore
    assert mapping.base_pair_dict[(g10, c18)].is_canonical  # type: ignore
    assert mapping.base_pair_dict[(a11, t17)].is_canonical  # type: ignore
    assert mapping.base_pair_dict[(a12, t16)].is_canonical  # type: ignore


# q-ugg is a quadruplex from molecular dynamics
def test_UGG_md():
    with open("tests/q-ugg-5k-salt_400-500ns_frame1065.pdb") as f:
        structure3d = read_3d_structure(f, 1)
    base_interactions = extract_base_interactions(structure3d)
    mapping = Mapping2D3D(
        structure3d, base_interactions.base_pairs, base_interactions.stackings, True
    )

    u1 = structure3d.find_residue(None, ResidueAuth(" ", 1, None, "U5"))
    g2 = structure3d.find_residue(None, ResidueAuth(" ", 2, None, "G"))
    g3 = structure3d.find_residue(None, ResidueAuth(" ", 3, None, "G"))
    u4 = structure3d.find_residue(None, ResidueAuth(" ", 4, None, "U"))
    g5 = structure3d.find_residue(None, ResidueAuth(" ", 5, None, "G"))
    g6 = structure3d.find_residue(None, ResidueAuth(" ", 6, None, "G"))

    u8 = structure3d.find_residue(None, ResidueAuth(" ", 8, None, "U5"))
    g9 = structure3d.find_residue(None, ResidueAuth(" ", 9, None, "G"))
    g10 = structure3d.find_residue(None, ResidueAuth(" ", 10, None, "G"))
    u11 = structure3d.find_residue(None, ResidueAuth(" ", 11, None, "U"))
    g12 = structure3d.find_residue(None, ResidueAuth(" ", 12, None, "G"))
    g13 = structure3d.find_residue(None, ResidueAuth(" ", 13, None, "G"))

    u15 = structure3d.find_residue(None, ResidueAuth(" ", 15, None, "U5"))
    g16 = structure3d.find_residue(None, ResidueAuth(" ", 16, None, "G"))
    g17 = structure3d.find_residue(None, ResidueAuth(" ", 17, None, "G"))
    u18 = structure3d.find_residue(None, ResidueAuth(" ", 18, None, "U"))
    g19 = structure3d.find_residue(None, ResidueAuth(" ", 19, None, "G"))
    g20 = structure3d.find_residue(None, ResidueAuth(" ", 20, None, "G"))

    u22 = structure3d.find_residue(None, ResidueAuth(" ", 22, None, "U5"))
    g23 = structure3d.find_residue(None, ResidueAuth(" ", 23, None, "G"))
    g24 = structure3d.find_residue(None, ResidueAuth(" ", 24, None, "G"))
    u25 = structure3d.find_residue(None, ResidueAuth(" ", 25, None, "U"))
    g26 = structure3d.find_residue(None, ResidueAuth(" ", 26, None, "G"))
    g27 = structure3d.find_residue(None, ResidueAuth(" ", 27, None, "G"))

    assert all(
        nt is not None
        for nt in [
            u1,
            g2,
            g3,
            u4,
            g5,
            g6,
            u8,
            g9,
            g10,
            u11,
            g12,
            g13,
            u15,
            g16,
            g17,
            u18,
            g19,
            g20,
            u22,
            g23,
            g24,
            u25,
            g26,
            g27,
        ]
    )

    # structure from MD have empty chain name
    assert u1.full_name == "U5/1"  # type: ignore
    assert g2.full_name == "G2"  # type: ignore

    # tetrad 1
    assert {u1, u15}.issubset(mapping.base_pair_graph[u8])  # type: ignore
    assert {u1, u15}.issubset(mapping.base_pair_graph[u22])  # type: ignore
    assert {u8, u22}.issubset(mapping.base_pair_graph[u1])  # type: ignore
    assert {u8, u22}.issubset(mapping.base_pair_graph[u15])  # type: ignore

    # tetrad 2
    assert {g2, g16}.issubset(mapping.base_pair_graph[g9])  # type: ignore
    assert {g2, g16}.issubset(mapping.base_pair_graph[g23])  # type: ignore
    assert {g9, g23}.issubset(mapping.base_pair_graph[g2])  # type: ignore
    assert {g9, g23}.issubset(mapping.base_pair_graph[g16])  # type: ignore

    # tetrad 3
    assert {g3, g17}.issubset(mapping.base_pair_graph[g10])  # type: ignore
    assert {g3, g17}.issubset(mapping.base_pair_graph[g24])  # type: ignore
    assert {g10, g24}.issubset(mapping.base_pair_graph[g3])  # type: ignore
    assert {g10, g24}.issubset(mapping.base_pair_graph[g17])  # type: ignore

    # tetrad 4
    assert {u4, u18}.issubset(mapping.base_pair_graph[u11])  # type: ignore
    assert {u4, u18}.issubset(mapping.base_pair_graph[u25])  # type: ignore
    assert {u11, u25}.issubset(mapping.base_pair_graph[u4])  # type: ignore
    assert {u11, u25}.issubset(mapping.base_pair_graph[u18])  # type: ignore

    # tetrad 5
    assert {g5, g19}.issubset(mapping.base_pair_graph[g12])  # type: ignore
    assert {g5, g19}.issubset(mapping.base_pair_graph[g26])  # type: ignore
    assert {g12, g26}.issubset(mapping.base_pair_graph[g5])  # type: ignore
    assert {g12, g26}.issubset(mapping.base_pair_graph[g19])  # type: ignore

    # tetrad 6
    assert {g6, g20}.issubset(mapping.base_pair_graph[g13])  # type: ignore
    assert {g6, g20}.issubset(mapping.base_pair_graph[g27])  # type: ignore
    assert {g13, g27}.issubset(mapping.base_pair_graph[g6])  # type: ignore
    assert {g13, g27}.issubset(mapping.base_pair_graph[g20])  # type: ignore



#! /usr/bin/env python
import argparse
import os
import tempfile
from collections import defaultdict, namedtuple
from typing import Iterable, List, Set, Tuple

from mmcif.io.IoAdapterPy import IoAdapterPy
from mmcif.io.PdbxReader import DataCategory, DataContainer

from rnapolis.util import handle_input_file

# Source: https://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Items/_entity_poly.type.html
ENTITY_POLY_TYPES = [
    "cyclic-pseudo-peptide",
    "other",
    "peptide nucleic acid",
    "polydeoxyribonucleotide",
    "polydeoxyribonucleotide/polyribonucleotide hybrid",
    "polypeptide(D)",
    "polypeptide(L)",
    "polyribonucleotide",
]

Link = namedtuple(
    "Link", ["parent_category_id", "parent_name", "child_category_id", "child_name"]
)


def load_pdbx_item_linked_group_list():
    dictionary = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "mmcif_pdbx_v50.dic"
    )
    adapter = IoAdapterPy()
    data = adapter.readFile(dictionary)
    obj = data[0].getObj("pdbx_item_linked_group_list")
    links = defaultdict(set)

    if obj:
        for row in obj.getRowList():
            row_dict = dict(zip(obj.getAttributeList(), row))
            child_category_id = row_dict["child_category_id"]
            child_name = row_dict["child_name"].split(".")[1]
            parent_name = row_dict["parent_name"].split(".")[1]
            parent_category_id = row_dict["parent_category_id"]
            links[parent_category_id].add(
                Link(parent_category_id, parent_name, child_category_id, child_name)
            )

    return links


def select_ids(
    data: List[DataContainer],
    category: str,
    field_name_to_extract: str,
    field_name_to_check: str,
    accepted_values: Iterable[str],
) -> Set[str]:
    obj = data[0].getObj(category)
    if not obj:
        return set()
    attributes = obj.getAttributeList()
    if field_name_to_check not in attributes or field_name_to_extract not in attributes:
        return set()
    index_to_check = attributes.index(field_name_to_check)
    index_to_extract = attributes.index(field_name_to_extract)
    return {
        row[index_to_extract]
        for row in obj.getRowList()
        if row[index_to_check] in accepted_values
    }


def select_category_by_id(
    data: List[DataContainer],
    category: str,
    field_name: str,
    ids: Iterable[str],
) -> Tuple[List[str], List[List[str]]]:
    obj = data[0].getObj(category)
    if not obj:
        return [], []
    attributes = obj.getAttributeList()
    if field_name not in attributes:
        return attributes, []
    index = attributes.index(field_name)
    return attributes, [row for row in obj.getRowList() if row[index] in ids]


def read_cif(file_content: str) -> DataContainer:
    with tempfile.NamedTemporaryFile("rt+") as f:
        adapter = IoAdapterPy()
        f.write(file_content)
        f.seek(0)
        return adapter.readFile(f.name)


def filter_cif(data, entity_ids, asym_ids, auth_asym_ids, retain_categories):
    links = load_pdbx_item_linked_group_list()
    categories_with_entity_id = [("entity", "id")] + [
        (link.child_category_id, link.child_name)
        for link in links["entity"]
        if link.parent_name == "id"
    ]
    categories_with_asym_id = [("struct_asym", "id")] + [
        (link.child_category_id, link.child_name)
        for link in links["struct_asym"]
        if link.parent_name == "id"
    ]
    categories_with_auth_asym_id = [("atom_site", "auth_asym_id")] + [
        (link.child_category_id, link.child_name)
        for link in links["atom_site"]
        if link.parent_name == "auth_asym_id"
    ]

    output = DataContainer("rnapolis")

    for table, ids in (
        (categories_with_entity_id, entity_ids),
        (categories_with_asym_id, asym_ids),
        (categories_with_auth_asym_id, auth_asym_ids),
    ):
        for category, field_name in table:
            attributes, rows = select_category_by_id(data, category, field_name, ids)

            if attributes and rows:
                obj = DataCategory(category, attributes, rows)
                output.append(obj)

    for category in retain_categories:
        obj = data[0].getObj(category)
        if obj:
            output.append(obj)

    with tempfile.NamedTemporaryFile("rt+") as tmp:
        adapter = IoAdapterPy()
        adapter.writeFile(tmp.name, [output])
        tmp.seek(0)
        return tmp.read()


def filter_by_poly_types(
    file_content: str,
    entity_poly_types: Iterable[str] = [
        "polyribonucleotide",
        "polydeoxyribonucleotide",
        "polydeoxyribonucleotide/polyribonucleotide hybrid",
    ],
    retain_categories: Iterable[str] = ["chem_comp"],
) -> str:
    data = read_cif(file_content)
    entity_ids = select_ids(
        data, "entity_poly", "entity_id", "type", set(entity_poly_types)
    )
    asym_ids = select_ids(data, "struct_asym", "id", "entity_id", entity_ids)
    auth_asym_ids = select_ids(
        data, "atom_site", "auth_asym_id", "label_asym_id", asym_ids
    )
    return filter_cif(data, entity_ids, asym_ids, auth_asym_ids, retain_categories)


def filter_by_chains(
    file_content: str,
    chains: Iterable[str],
    retain_categories: Iterable[str] = ["chem_comp"],
) -> str:
    """
    Filter a PDBx/mmCIF file by chain IDs. The function returns a new PDBx/mmCIF file.

    Warning! The new file might contain more chains than provided in the `chains` argument.
    This is because the function filters by entity, so if you ask for chain "A",
    which is part of entity 1 having chains "A", "B" and "C", then you will get all three chains.
    """
    data = read_cif(file_content)
    asym_ids = set(chains)
    entity_ids = select_ids(data, "struct_asym", "entity_id", "id", asym_ids)
    auth_asym_ids = select_ids(
        data, "atom_site", "auth_asym_id", "label_asym_id", asym_ids
    )
    return filter_cif(data, entity_ids, asym_ids, auth_asym_ids, retain_categories)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--filter-by-poly-types",
        help=f"filter by entity poly types, possible values: {', '.join(ENTITY_POLY_TYPES)}",
        action="append",
        default=[],
    )
    parser.add_argument(
        "--filter-by-chains",
        help="filter by chain IDs (label_asym_id), e.g. A, B, C",
        action="append",
        default=[],
    )
    parser.add_argument(
        "--retain-categories",
        help="categories to retain in the output file default: chem_comp",
        action="append",
        default=["chem_comp"],
    )
    parser.add_argument("path", help="path to a PDBx/mmCIF file")
    args = parser.parse_args()

    file = handle_input_file(args.path)
    if args.filter_by_poly_types:
        print(
            filter_by_poly_types(
                file.read(),
                entity_poly_types=args.filter_by_poly_types,
                retain_categories=args.retain_categories,
            )
        )
    elif args.filter_by_chains:
        print(
            filter_by_chains(
                file.read(),
                chains=args.filter_by_chains,
                retain_categories=args.retain_categories,
            )
        )
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



from pathlib import Path

from rnapolis.adapter import ExternalTool, process_external_tool_output
from rnapolis.parser import read_3d_structure
from rnapolis.util import handle_input_file


def test_adapter_fr3d():
    """Test adapter with FR3D output for 1A4D structure."""
    test_dir = Path(__file__).parent
    cif_path = test_dir / "1A4D_1_A-B.cif"
    fr3d_path = test_dir / "1A4D_1_A-B-basepair_detail.txt"

    # Read 3D structure
    file = handle_input_file(str(cif_path))
    structure3d = read_3d_structure(file, None)

    # Process external tool output
    structure2d, mapping = process_external_tool_output(
        structure3d,
        [str(fr3d_path)],
        ExternalTool.FR3D,
        str(cif_path),
        find_gaps=False,
    )

    # Check the dot-bracket output
    expected_dot_bracket = """>strand_A
GGCCGAUGGUAGUGUGGGGUC
((((.......(((((((...
>strand_B
UCCCCAUGCGAGAGUAGGCC
..))))))).......))))"""

    assert mapping.dot_bracket == expected_dot_bracket


def test_adapter_bpnet():
    """Test adapter with BPNet output for 1A4D structure."""
    test_dir = Path(__file__).parent
    cif_path = test_dir / "1A4D_1_A-B.cif"
    bpnet_path = test_dir / "1A4D_1_A-B-input_basepair.json"

    # Read 3D structure
    file = handle_input_file(str(cif_path))
    structure3d = read_3d_structure(file, None)

    # Process external tool output
    structure2d, mapping = process_external_tool_output(
        structure3d,
        [str(bpnet_path)],
        ExternalTool.BPNET,
        str(cif_path),
        find_gaps=False,
    )

    # Check the dot-bracket output
    expected_dot_bracket = """>strand_A
GGCCGAUGGUAGUGUGGGGUC
((((.......(((((((...
>strand_B
UCCCCAUGCGAGAGUAGGCC
..))))))).......))))"""

    assert mapping.dot_bracket == expected_dot_bracket


def test_adapter_rnaview():
    """Test adapter with RNAView output for 1A4D structure."""
    test_dir = Path(__file__).parent
    cif_path = test_dir / "1A4D_1_A-B.cif"
    rnaview_path = test_dir / "1A4D_1_A-B-input.cif.out"

    # Read 3D structure
    file = handle_input_file(str(cif_path))
    structure3d = read_3d_structure(file, None)

    # Process external tool output
    structure2d, mapping = process_external_tool_output(
        structure3d,
        [str(rnaview_path)],
        ExternalTool.RNAVIEW,
        str(cif_path),
        find_gaps=False,
    )

    # Check the dot-bracket output
    expected_dot_bracket = """>strand_A
GGCCGAUGGUAGUGUGGGGUC
((((.......(((((((...
>strand_B
UCCCCAUGCGAGAGUAGGCC
..))))))).......))))"""

    assert mapping.dot_bracket == expected_dot_bracket


def test_adapter_maxit():
    """Test adapter with MAXIT output for 1A4D structure."""
    test_dir = Path(__file__).parent
    cif_path = test_dir / "1A4D_1_A-B.cif"
    maxit_path = test_dir / "1A4D_1_A-B-output.cif"

    # Read 3D structure
    file = handle_input_file(str(cif_path))
    structure3d = read_3d_structure(file, None)

    # Process external tool output
    structure2d, mapping = process_external_tool_output(
        structure3d,
        [str(maxit_path)],
        ExternalTool.MAXIT,
        str(cif_path),
        find_gaps=False,
    )

    # Check the dot-bracket output
    expected_dot_bracket = """>strand_A
GGCCGAUGGUAGUGUGGGGUC
((((.......((((((((..
>strand_B
UCCCCAUGCGAGAGUAGGCC
.)))))))).......))))"""

    assert mapping.dot_bracket == expected_dot_bracket


def test_adapter_dssr():
    """Test adapter with DSSR output for 1A4D structure."""
    test_dir = Path(__file__).parent
    cif_path = test_dir / "1A4D_1_A-B.cif"
    dssr_path = test_dir / "1A4D_1_A-B-dssr.json"

    # Read 3D structure
    file = handle_input_file(str(cif_path))
    structure3d = read_3d_structure(file, None)

    # Process external tool output
    structure2d, mapping = process_external_tool_output(
        structure3d,
        [str(dssr_path)],
        ExternalTool.DSSR,
        str(cif_path),
        find_gaps=False,
    )

    # Check the dot-bracket output
    expected_dot_bracket = """>strand_A
GGCCGAUGGUAGUGUGGGGUC
((((.......((((((((..
>strand_B
UCCCCAUGCGAGAGUAGGCC
.)))))))).......))))"""

    assert mapping.dot_bracket == expected_dot_bracket


def test_adapter_auto_detection():
    """Test that auto-detection works correctly for different file types."""
    test_dir = Path(__file__).parent
    cif_path = test_dir / "1A4D_1_A-B.cif"

    # Read 3D structure once
    file = handle_input_file(str(cif_path))
    structure3d = read_3d_structure(file, None)

    # Test auto-detection for each tool
    test_cases = [
        (test_dir / "1A4D_1_A-B-basepair_detail.txt", ExternalTool.FR3D),
        (test_dir / "1A4D_1_A-B-input_basepair.json", ExternalTool.BPNET),
        (test_dir / "1A4D_1_A-B-input.cif.out", ExternalTool.RNAVIEW),
        (test_dir / "1A4D_1_A-B-output.cif", ExternalTool.MAXIT),
        (test_dir / "1A4D_1_A-B-dssr.json", ExternalTool.DSSR),
    ]

    from rnapolis.adapter import auto_detect_tool

    for file_path, expected_tool in test_cases:
        detected_tool = auto_detect_tool([str(file_path)])
        assert detected_tool == expected_tool, (
            f"Auto-detection failed for {file_path}: expected {expected_tool}, got {detected_tool}"
        )


def test_adapter_empty_files_maxit():
    """Test adapter with empty external files (should default to MAXIT)."""
    test_dir = Path(__file__).parent
    cif_path = test_dir / "1A4D_1_A-B.cif"

    # Read 3D structure
    file = handle_input_file(str(cif_path))
    structure3d = read_3d_structure(file, None)

    # Process with empty external files (should use MAXIT with input file)
    structure2d, mapping = process_external_tool_output(
        structure3d,
        [],  # Empty external files
        ExternalTool.MAXIT,
        str(cif_path),
        find_gaps=False,
    )

    # Should still produce some output (even if empty interactions)
    assert mapping.dot_bracket is not None
    assert structure2d is not None



#!/usr/bin/env python3
import argparse
import os
import tempfile

import pandas as pd

from rnapolis.parser import is_cif
from rnapolis.parser_v2 import parse_cif_atoms, parse_pdb_atoms, write_cif, write_pdb
from rnapolis.tertiary_v2 import Structure


def main():
    """Main function to run the unifier tool."""
    parser = argparse.ArgumentParser(description="Align two PDB or mmCIF files.")
    parser.add_argument("--output", "-o", help="Output directory", required=True)
    parser.add_argument(
        "--format",
        "-f",
        help="Output format (possible values: PDB, mmCIF, keep. Default: keep)",
        default="keep",
    )
    parser.add_argument("pdb1", help="First PDB or mmCIF file")
    parser.add_argument("pdb2", help="Second PDB or mmCIF file")
    args = parser.parse_args()

    from pymol import cmd

    cmd.load(args.pdb1, "pdb1")
    cmd.load(args.pdb2, "pdb2")
    cmd.align("pdb1", "pdb2", object="aligned", cycles=0)

    pdb1_aligned = []
    pdb2_aligned = []

    with tempfile.NamedTemporaryFile("wt+", suffix=".aln") as f:
        cmd.save(f.name, "aligned")
        f.seek(0)

        for line in f:
            if line.startswith("pdb1"):
                pdb1_aligned.append(line.split()[1])
            elif line.startswith("pdb2"):
                pdb2_aligned.append(line.split()[1])

    pdb1_aligned = "".join(pdb1_aligned)
    pdb2_aligned = "".join(pdb2_aligned)
    residues_to_remove = {"pdb1": [], "pdb2": []}

    i, j = 0, 0
    for c1, c2 in zip(pdb1_aligned, pdb2_aligned):
        if c1 == c2 == "-":
            continue  # Should not happen to have gap aligned to gap, but just in case

        if c1 == c2:
            i += 1
            j += 1
            continue

        if c1 == "-":
            residues_to_remove["pdb2"].append(j)
            j += 1
            continue

        if c2 == "-":
            residues_to_remove["pdb1"].append(i)
            i += 1
            continue

        if c1 != c2:
            residues_to_remove["pdb1"].append(i)
            residues_to_remove["pdb2"].append(j)
            i += 1
            j += 1
            continue

        raise ValueError("This should not happen!")

    if not residues_to_remove["pdb1"] and not residues_to_remove["pdb2"]:
        print("Structures are already aligned")

    structures = {}
    for key, path in [("pdb1", args.pdb1), ("pdb2", args.pdb2)]:
        with open(path) as f:
            if is_cif(f):
                atoms = parse_cif_atoms(f)
            else:
                atoms = parse_pdb_atoms(f)

        structures[key] = Structure(atoms).residues

    for key, residues in structures.items():
        for i in sorted(residues_to_remove[key], reverse=True):
            del residues[i]

    # Write output
    os.makedirs(args.output, exist_ok=True)

    for (key, residues), path in zip(structures.items(), [args.pdb1, args.pdb2]):
        base, _ = os.path.splitext(os.path.basename(path))

        if args.format == "keep":
            format = residues[0].atoms.attrs["format"]
        else:
            format = args.format

        ext = ".pdb" if format == "PDB" else ".cif"

        with open(f"{args.output}/{base}{ext}", "w") as f:
            df = pd.concat([residue.atoms for residue in residues])

            if format == "PDB":
                write_pdb(df, f)
            else:
                write_cif(df, f)


if __name__ == "__main__":
    main()



from setuptools import setup

with open("README.md") as f:
    long_description = f.read()

setup(
    name="RNApolis",
    version="0.10.0",
    packages=["rnapolis"],
    package_dir={"": "src"},
    author="Tomasz Zok",
    author_email="tomasz.zok@cs.put.poznan.pl",
    description="A Python library containing RNA-related bioinformatics functions and classes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tzok/rnapolis-py",
    project_urls={"Bug Tracker": "https://github.com/tzok/rnapolis-py/issues"},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    entry_points={
        "console_scripts": [
            "annotator=rnapolis.annotator:main",
            "adapter=rnapolis.adapter:main",
            "aligner=rnapolis.aligner:main",
            "clashfinder=rnapolis.clashfinder:main",
            "distiller=rnapolis.distiller:main",
            "metareader=rnapolis.metareader:main",
            "molecule-filter=rnapolis.molecule_filter:main",
            "motif-extractor=rnapolis.motif_extractor:main",
            "transformer=rnapolis.transformer:main",
            "rfam-folder=rnapolis.rfam_folder:main",
            "unifier=rnapolis.unifier:main",
            "splitter=rnapolis.splitter:main",
        ]
    },
    include_package_data=True,
    package_data={
        "rnapolis": [
            "component_A.csv",
            "component_C.csv",
            "component_G.csv",
            "component_U.csv",
            "mmcif_pdbx_v50.dic",
        ]
    },
    install_requires=[
        "appdirs",
        "faiss-cpu",
        "graphviz",
        "mmcif",
        "numpy",
        "ordered-set",
        "orjson",
        "pandas",
        "pulp",
        "requests",
        "scikit-learn",
        "scipy",
        "tqdm",
        "viennarna",
    ],
)



import tempfile

from rnapolis.parser import parse_cif
from rnapolis.transformer import copy_from_to, replace_value


def test_replace_value():
    with open("tests/4gqj-assembly1.cif") as f:
        content = f.read()

    with tempfile.NamedTemporaryFile(mode="wt") as f:
        f.write(content)
        f.seek(0)
        org_atoms, _, _, _ = parse_cif(f)

    org_label_asym_id = set([atom.label.chain for atom in org_atoms if atom.label])
    org_auth_asym_id = set([atom.auth.chain for atom in org_atoms if atom.auth])
    assert org_label_asym_id == set(["A", "B", "A-2", "B-2"])
    assert org_auth_asym_id == set(["A", "B", "A-2", "B-2"])

    replaced_content, mapping = replace_value(
        content, "atom_site", "auth_asym_id", "ABCD"
    )
    assert mapping == {"A": "A", "B": "B", "A-2": "C", "B-2": "D"}

    with tempfile.NamedTemporaryFile(mode="rt+") as f:
        f.write(replaced_content)
        f.seek(0)
        rep_atoms, _, _, _ = parse_cif(f)

    rep_label_asym_id = set([atom.label.chain for atom in rep_atoms if atom.label])
    rep_auth_asym_id = set([atom.auth.chain for atom in rep_atoms if atom.auth])
    assert rep_label_asym_id == set(["A", "B", "A-2", "B-2"])
    assert rep_auth_asym_id == set(["A", "B", "C", "D"])


def test_copy_from_to():
    with open("tests/5it9.cif") as f:
        content = f.read()

    with tempfile.NamedTemporaryFile(mode="wt") as f:
        f.write(content)
        f.seek(0)
        org_atoms, _, _, _ = parse_cif(f)

    org_label_asym_id = set([atom.label.chain for atom in org_atoms if atom.label])
    org_auth_asym_id = set([atom.auth.chain for atom in org_atoms if atom.auth])
    assert org_label_asym_id == set(["HA", "IA"])
    assert org_auth_asym_id == set(["2", "i"])

    replaced_content = copy_from_to(
        content, "atom_site", "label_asym_id", "auth_asym_id"
    )

    with tempfile.NamedTemporaryFile(mode="rt+") as f:
        f.write(replaced_content)
        f.seek(0)
        rep_atoms, _, _, _ = parse_cif(f)

    rep_label_asym_id = set([atom.label.chain for atom in rep_atoms if atom.label])
    rep_auth_asym_id = set([atom.auth.chain for atom in rep_atoms if atom.auth])
    assert rep_label_asym_id == set(["HA", "IA"])
    assert rep_auth_asym_id == set(["HA", "IA"])



from rnapolis.annotator import extract_base_interactions
from rnapolis.common import ResidueAuth, ResidueLabel
from rnapolis.parser import read_3d_structure
from rnapolis.tertiary import Mapping2D3D


# in 1E7K there is a break in a chain, so it should be recognized as separate strands
def test_1E7K():
    with open("tests/1E7K_1_C.cif") as f:
        structure3d = read_3d_structure(f, 1)
    base_interaction = extract_base_interactions(structure3d)
    mapping = Mapping2D3D(
        structure3d, base_interaction.base_pairs, base_interaction.stackings, True
    )
    assert len(mapping.strands_sequences) == 1


# in 1DFU the adjacent U and G seem to be base-pair like if you do not take into account angles
def test_1DFU():
    with open("tests/1DFU_1_M-N.cif") as f:
        structure3d = read_3d_structure(f, 1)

    b1u = structure3d.find_residue(ResidueLabel("B", 1, "U"), None)
    assert b1u is not None

    b2g = structure3d.find_residue(ResidueLabel("B", 2, "G"), None)
    assert b2g is not None

    base_interactions = extract_base_interactions(structure3d)
    mapping = Mapping2D3D(
        structure3d, base_interactions.base_pairs, base_interactions.stackings, True
    )
    assert b2g not in mapping.base_pair_graph[b1u]
    assert b1u not in mapping.base_pair_graph[b2g]


# in 4WTI the first residue has only O3' atom, but is stil considered a nucleotide
def test_4WTI():
    with open("tests/4WTI_1_T-P.cif") as f:
        structure3d = read_3d_structure(f, 1)
    base_interactions = extract_base_interactions(structure3d)
    mapping = Mapping2D3D(
        structure3d, base_interactions.base_pairs, base_interactions.stackings, True
    )
    assert mapping.dot_bracket == ">strand_T\nCGG\n.((\n>strand_P\nCC\n))"


# in 1HMH the bases are oriented in 45 degrees and it caused the program to identify invalid base pair
def test_1HMH():
    with open("tests/1HMH_1_E.cif") as f:
        structure3d = read_3d_structure(f, 1)
    base_interactions = extract_base_interactions(structure3d)
    mapping = Mapping2D3D(
        structure3d, base_interactions.base_pairs, base_interactions.stackings, True
    )
    assert mapping.dot_bracket == ">strand_E\nUG\n.."


# in 6INQ the residues T.DC0 and N.DG0 were not found by RNApolis
def test_6INQ():
    with open("tests/6INQ.cif") as f:
        structure3d = read_3d_structure(f, 1)
    assert structure3d.find_residue(ResidueLabel("N", 73, "DC"), None) is not None
    assert structure3d.find_residue(None, ResidueAuth("T", 0, None, "DC")) is not None
    assert structure3d.find_residue(ResidueLabel("O", 126, "DG"), None) is not None
    assert structure3d.find_residue(None, ResidueAuth("N", 0, None, "DG")) is not None


# in 6g90 from rna3db, the sequence contains Ns which were ignored incorrectly
def test_6g90():
    with open("tests/6g90_1.cif") as f:
        structure3d = read_3d_structure(f, nucleic_acid_only=True)
    sequence = "".join([residue.one_letter_name for residue in structure3d.residues])
    assert (
        sequence
        == "AUACUUACCUUAAGAUAUCAGAGGAGAUCAAGAAGUCCUACUGAUCAAACAUGCGCUUCCAAGAAGGACGUUAAGCAUUUAUCAUUGAACGUUCAUUGAACAUUGAUGCAAACUCCUUGGUCACACACACGCGGAAGGCGUGUUUGCUGACGUCCCUUGUUUCAAUCAUUGGUUAACUGAUUUUUGGGGCCCUUUGUUCUUCUGAGAAGUGACACCAAUUGGUGUUAGGGGAGCUGGGGCCUUUCAAAANNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNUUUUGGAAGGUCUUGGUCGGGUGGAUCUUAUAAUUUUUGAUUUA"
    )

    base_interactions = extract_base_interactions(structure3d)

    # without gaps
    mapping = Mapping2D3D(
        structure3d, base_interactions.base_pairs, base_interactions.stackings, False
    )
    assert (
        mapping.bpseq.sequence
        == mapping.dot_bracket.split("\n")[1]
        == "AUACUUACCUUAAGAUAUCAGAGGAGAUCAAGAAGUCCUACUGAUCAAACAUGCGCUUCCAAGAAGGACGUUAAGCAUUUAUCAUUGAACGUUCAUUGAACAUUGAUGCAAACUCCUUGGUCACACACACGCGGAAGGCGUGUUUGCUGACGUCCCUUGUUUCAAUCAUUGGUUAACUGAUUUUUGGGGCCCUUUGUUCUUCUGAGAAGUGACACCAAUUGGUGUUAGGGGAGCUGGGGCCUUUCAAAANNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNUUUUGGAAGGUCUUGGUCGGGUGGAUCUUAUAAUUUUUGAUUUA"
    )

    # with gaps
    mapping = Mapping2D3D(
        structure3d, base_interactions.base_pairs, base_interactions.stackings, True
    )
    assert (
        mapping.bpseq.sequence
        == mapping.dot_bracket.split("\n")[1]
        == "AUACUUACCUUAAGAUAUCAGAGGAGAUCAAGAAGUCCUACUGAUCAAACAUGCGCUUCCA?????AGAAGGACGUUAAGCAUUUAUCAUUGAAC???????GUUCAUUGAA??CAUUGAUGCAAACUCCUUGGUCACACACAC???????GCGGAAGGCGUGUUUGCUGACG???????UCCCUUGUUUCAAUCAUUGGUU?????????????????????????????????AACUGAUUUUUGGGGCCCUUUGUU?CUUCUG?????AGAAGU??GACACCAA???????UUGGUGUUAGGGGAGCUGGGGCCUUUCAAAANNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNUUUUGGAAGGUCUUGGUCGGGUGGAUCUUAUAAUUUUUGAUUUA"
    )
